__title__ = "phoenyx"
__description__ = "A drawing and physics engine for Pygame"
__url__ = "https://github.com/ThomasByr/Phoenyx"
__version__ = "0.3.3"
__author__ = "Thomas Byr"
__author_email__ = "thomas-c2000@outlook.com"
__license__ = "GNU GPLv3"
# __copyright__ = "Copyright (c) 2021 Thomas Byr"



from typing import Union

from . import Renderer

import pymunk
import pymunk.pygame_util

__all__ = ["SandBox"]

from ..data import *
from ..pmath import *
from ..elements import *


def _map(x: float, x0: float, x1: float, y0: float, y1: float) -> float:
    """
    linear interpolation
    """
    return (y0 * (x1-x) + y1 * (x-x0)) / (x1-x0)


def _constrain(x: float, mn: float, mx: float) -> int:
    """
    value constrain
    """
    return mn if x < mn else mx if x > mx else x


class SandBox:
    """
    Phoenyx SandBox
    ===============
    Provides :
    1. basic physics for simple shapes
    2. fully integrated to Phoenyx Renderer
    3. C acceleration
    """
    def __init__(self,
                 renderer: Renderer,
                 width: int = None,
                 height: int = None,
                 bounce: bool = False) -> None:
        """
        new SandBox instance

        Parameters
        ----------
            renderer : Renderer
                main renderer
            width : int, (optional)
                width of the world from the center
                defaults to None
            height : int, (optional)
                height of the world from the center
                defaults to None
            bounce : bool, (optional)
                if bodies bounce on the edges of the world
                defaults to False

        Note
        ----
            The center of the SandBox is the center of the Renderer window ;\\
            The default size of the SandBox is set to fill the Renderer window ;\\
            The default gravitational constant is set to 900 downwards.
        """
        self._renderer = renderer

        self._buffer = 10
        self._bounce = bounce

        self._x, self._y = self._renderer.win_width / 2, self._renderer.win_height / 2
        self._width = width if width is not None else self._renderer.win_width / 2
        self._height = height if height is not None else self._renderer.win_height / 2

        self._sum_of_forces = Vector()
        self._gravity = Vector(0, 900)

        self._borders: set[pymunk.Shape] = set()
        self._all_shapes: set[pymunk.Shape] = set()
        self._space = pymunk.Space()
        self._space.gravity = self._gravity.x, self._gravity.y

        self._draw_options = pymunk.pygame_util.DrawOptions(renderer._window)

        if bounce:
            self._add_borders()

    @property
    def width(self) -> int:
        """
        gets current width from center
        """
        return self._width

    @property
    def height(self) -> int:
        """
        gets current height from center
        """
        return self._height

    @property
    def bodies(self) -> list[pymunk.Body]:
        """
        gets current living bodies
        """
        return self._space.bodies

    @property
    def shapes(self) -> list[pymunk.Shape]:
        """
        gets current living shapes
        """
        return self._space.shapes

    def _add_borders(self) -> None:
        o = -self._buffer
        w = 2 * self.width - o
        h = 2 * self.height - o
        s1 = pymunk.Segment(self._space.static_body, (o, o), (w, o), -o - 1)
        s1.friction = .99
        s1.elasticity = .99
        s2 = pymunk.Segment(self._space.static_body, (w, o), (w, h), -o - 1)
        s2.friction = .99
        s2.elasticity = .99
        s3 = pymunk.Segment(self._space.static_body, (w, h), (o, h), -o - 1)
        s3.friction = .99
        s3.elasticity = .99
        s4 = pymunk.Segment(self._space.static_body, (o, h), (o, o), -o - 1)
        s4.friction = .99
        s4.elasticity = .99

        self._borders.add(s1)
        self._borders.add(s2)
        self._borders.add(s3)
        self._borders.add(s4)
        self._space.add(s1)
        self._space.add(s2)
        self._space.add(s3)
        self._space.add(s4)

    def set_bouncing(self, bounce: bool) -> None:
        """
        sets the bouncing behavior of the bodies

        Parameters
        ----------
            bounce : bool
                if bodies bounce on the edges of the world
        """
        warn(
            f"WARNING [sandbox] : change in bodies bouncing behavior, may alter simulation"
        )
        if self._wrap and bounce:
            warn(
                f"ERROR [sandbox] : bouncing and wraping can not be both active, nothing changed"
            )
            return
        self._bounce = bounce

        if bounce:
            self._add_borders()

        else:
            for wall in self._borders:
                self._space.remove(wall)
            self._borders.clear()

    def set_gravity(self, x: float = 0, y: float = 0) -> None:
        """
        sets global gravity\\
        note that gravity affects all objects that have a mass\\
        but does not depend on that mass assuming it is not equal to zero

        Parameters
        ----------
            x : float, (optional)
                x component of the g vector ; (x > 0 is pointing to the right)
            y : float, (optional)
                y component of the g vector ; (y > 0 is pointing down)
        """
        self._gravity = Vector(x, y)
        self._space.gravity = self._gravity.x, self._gravity.y

    def _is_out(self, position: tuple[float, float]) -> bool:
        x, y = position
        w = 10 * self.width + self._buffer
        h = 10 * self.height + self._buffer
        return not ((self._x - w <= x <= self._x + w)\
               and (self._y - h <= y <= self._y + h))

    def _get_center(self, *points: tuple[int, int]) -> tuple[int, int]:
        """
        
        """
        cx = 0
        cy = 0
        for p in points:
            cx += p[0]
            cy += p[1]
        x, y = cx / len(points), cy / len(points)

        return x, y

    def add_ball(self,
                 x: float,
                 y: float,
                 mass: float,
                 radius: int,
                 friction: float = .99,
                 elasticity: float = 0,
                 is_static: bool = False) -> pymunk.Circle:
        """
        new circular body with uniform mass repartition

        Parameters
        ----------
            x : float
                x location of the Body
            y : float
                y location of the Body
            mass : float
                mass of the Body
            radius : float
                outer radius of the circle

        Options
        -------
            fiction : float, (optional)
                defaults to .99
            elasticity : float, (optional)
                defaults to 0
            is_static : bool, (optional)
                defaults to False
        """
        inertia = pymunk.moment_for_circle(mass, 0, radius, (0, 0))

        opt = {
            "body_type":
            pymunk.Body.STATIC if is_static else pymunk.Body.DYNAMIC
        }
        body = pymunk.Body(mass, inertia, **opt)
        body.position = x, y

        shape = pymunk.Circle(body, radius, (0, 0))
        shape.friction = friction
        shape.elasticity = elasticity

        self._space.add(body, shape)
        self._all_shapes.add(shape)
        return shape

    def add_segment(self,
                    p1: Union[tuple[float, float], Vector],
                    p2: Union[tuple[float, float], Vector],
                    mass: float,
                    radius: float,
                    friction: float = .99,
                    elasticity: float = 0,
                    is_static: bool = False) -> pymunk.Segment:
        """
        new static Segment body with uniform mass repartition

        Parameters
        ----------
            p1 : Union[tuple[float, float], Vector]
                position of the first vertex
            p2 : Union[tuple[float, float], Vector]
                position of the second vertex
            mass: float
                mass of segment
            radius : float
                radius of segment

        Options
        -------
            fiction : float, (optional)
                defaults to .99
            elasticity : float, (optional)
                defaults to 0
            is_static : bool, (optional)
                defaults to False
        """
        a = p1[0], p1[1]
        b = p2[0], p2[1]

        inertia = pymunk.moment_for_segment(mass, a, b, radius)
        opt = {
            "body_type":
            pymunk.Body.STATIC if is_static else pymunk.Body.DYNAMIC
        }
        body = pymunk.Body(mass, inertia, **opt)

        shape = pymunk.Segment(body, a, b, radius)
        shape.friction = friction
        shape.elasticity = elasticity

        if not is_static:
            body.position = self._get_center(a, b)
        body.center_of_gravity = shape.center_of_gravity

        self._space.add(body, shape)
        self._all_shapes.add(shape)
        return shape

    def add_poly(self,
                 points: list[Union[tuple[int, int], Vector]],
                 mass: float,
                 radius: float = .01,
                 friction: float = .99,
                 elasticity: float = 0,
                 is_static: bool = False) -> pymunk.Poly:
        """
        new convex Polygon body with uniform mass repartition

        Parameters
        ----------
            points : list[Union[tuple[int, int], Vector]]
                position of the vertexes
            mass: float
                mass of polygon

        Options
        -------
            radius : float, (optional)
                defaults to .01
            fiction : float, (optional)
                defaults to .99
            elasticity : float, (optional)
                defaults to 0
            is_static : bool, (optional)
                defaults to False

        Note
        ----
            adding a small radius bevel the corners and can significantly reduce problems where the poly gets stuck on seams in your geometry
        """
        points = [(p[0], p[1]) for p in points]

        inertia = pymunk.moment_for_poly(mass, points, radius=radius)
        opt = {
            "body_type":
            pymunk.Body.STATIC if is_static else pymunk.Body.DYNAMIC
        }
        body = pymunk.Body(mass, inertia, **opt)

        shape = pymunk.Poly(body, points, radius=radius)
        shape.friction = friction
        shape.elasticity = elasticity

        body.position = self._get_center(*points)
        body.center_of_gravity = shape.center_of_gravity

        self._space.add(body, shape)
        self._all_shapes.add(shape)
        return shape

    def extend_segment(self,
                       segment: pymunk.Segment,
                       pos: Union[tuple[float, float], Vector],
                       angle: float,
                       len: float,
                       mass: float,
                       radius: float,
                       friction: float = .99,
                       elasticity: float = 0) -> pymunk.Segment:
        """
        extends an existing segment

        Parameters
        ----------
            segment : pymunk.Segment
                the segment to extend
            pos : Union[tuple[float, float], Vector]
                base position
            angle : float
                the angle
            len : float
                the length
            mass : float
                mass of segment to add
            radius : float
                radius of segment to add

        Options
        -------
            friction : float, (optional)
                defaults to .99
            elasticity : float, (optional)
                defaults to 0

        Note
        ----
            note that extending a dynamic segment may introduce a transient state\\
            also, the static or dynamic nature will follow the base segment
        """
        body = segment.body

        a = pos[0], pos[1]
        p1 = Vector(pos[0], pos[1])
        v = Vector(len)
        v.angle = angle
        p2 = p1 + v
        b = p2.x, p2.y

        seg = pymunk.Segment(body, a, b, radius)
        seg.mass = mass
        seg.friction = friction
        seg.elasticity = elasticity

        self._space.add(seg)
        self._all_shapes.add(seg)
        return seg

    def add_pin_joint(self, pos: Union[tuple[float, float], Vector],
                      shape: pymunk.Shape) -> pymunk.PinJoint:
        """
        new static pin joint

        Parameters
        ----------
            pos : Union[tuple[float, float], Vector]
                position of the pin joint
            shape : Shape
                the shape to attach to\\
                can be Circle, Segment and Poly
        """
        pos = pos[0], pos[1]
        rotation_center_body = pymunk.Body(body_type=pymunk.Body.STATIC)
        rotation_center_body.position = pos
        body = shape.body
        # body.position = pos

        rotation_center_joint = pymunk.PinJoint(body, rotation_center_body,
                                                pos, (0, 0))
        rotation_center_joint.distance = 0
        self._space.add(rotation_center_joint)
        return rotation_center_joint

    def add_slide_joint(
            self, pos: Union[tuple[float, float], Vector], shape: pymunk.Shape,
            limit: Union[float, tuple[float, float]]) -> pymunk.SlideJoint:
        """
        new static slide joint

        Parameters
        ----------
            pos : Union[tuple[float, float], Vector]
                position of the slide joint
            shape : pymunk.Shape
                the shape to attach to\\
                can be Circle, Segment and Poly
            limit : Union[float, tuple[float, float]]
                 max distance limit, optional lower limit
        """
        pos = pos[0], pos[1]
        rotation_limit_body = pymunk.Body(body_type=pymunk.Body.STATIC)
        rotation_limit_body.position = pos
        body = shape.body

        up = limit if isinstance(limit, (float, int)) else max(limit)
        low = 0 if isinstance(limit, (float, int)) else min(limit)
        rotation_limit_joint = pymunk.SlideJoint(body, rotation_limit_body,
                                                 pos, (0, 0), low, up)
        self._space.add(rotation_limit_joint)
        return rotation_limit_joint

    def clear(self) -> None:
        """
        clear space : will delete all shapes and bodies
        """
        to_remove: set[Union[pymunk.Body, pymunk.Shape]] = set()
        for b in self._space.bodies:
            to_remove.add(b)
        for s in self._space.shapes:
            to_remove.add(s)

        for e in to_remove:
            self._space.remove(e)
            self._all_shapes.discard(e)

    def discard(self, shape: pymunk.Shape) -> None:
        """
        removes a shape and its body from the space\\
        won't raise any errors, hopefully
        """
        try:
            self._space.remove(shape)
        except AssertionError:
            pass
        try:
            self._space.remove(shape.body)
        except AssertionError:
            pass
        self._all_shapes.discard(shape)

    def step(self, fps: int = 60, iter: int = 10) -> None:
        """
        go forward in time by one step\\
        the dt used for computation is based on the parameters

        Parameters
        ----------
            fps : int, (optional)
                number of frames per second
                defaults to 60
            iter : int, (optional)
                number of iterations to perform, could increase accuracy
                defaults to 10
        """
        dt = 1 / (fps*iter)
        for _ in range(iter):
            self._space.step(dt)

        shapes_to_remove: set[pymunk.Shape] = set()
        for s in self._all_shapes:
            if self._is_out(s.body.position):
                shapes_to_remove.add(s)

        for s in shapes_to_remove:
            self._space.remove(s, s.body)
            self._all_shapes.discard(s)

    def draw(self) -> None:
        """
        default drawing method for the physics engine\\
        usefull for debuging
        """
        self._space.debug_draw(self._draw_options)



from typing import Union
from ..data import *

__all__ = ["Slider"]

import difflib
import pygame


def _map(x: float, x0: float, x1: float, y0: float, y1: float) -> float:
    """
    linear interpolation
    """
    return (y0 * (x1-x) + y1 * (x-x0)) / (x1-x0)


class Slider:
    """
    Pygame Slider
    =============
    created by ``Renderer``
    """
    def __init__(self,
                 renderer,
                 x: int,
                 y: int,
                 name: str,
                 min_val: float,
                 max_val: float,
                 value: float,
                 incr: int,
                 radius: int = 7,
                 shape: str = CIRCLE,
                 thickness: int = 3,
                 color: Union[tuple[int, int, int], int,
                              str] = (155, 155, 155),
                 fullcolor: Union[tuple[int, int, int], int,
                                  str] = (155, 70, 70),
                 length: int = 100,
                 count: int = 30) -> None:
        """
        new slider instance

        Parameters
        ----------
            renderer : Renderer
                the Renderer instance the Slider is linked to
            x : int
                x-coordinate
            y : int
                y-coordinate
            name : str
                name of the slider (must be unique !)
            min_val : float
                minimum value
            max_val : float
                maximum value
            value : float
                current value
            incr : int
                number of digits
            radius : int, (optional)
                radius of the slider cursor
                defaults to 7
            shape : str, (optional)
                shape of the slider cursor
                SQUARE | CIRCLE | CROSS | PLUS
                defaults to CIRCLE
            thickness : int, (optional)
                thickness of the slider bar
                defaults to 3
            color : Union[tuple[int, int, int], int, str], (optional)
                default color of the bar
                defaults to (155, 155, 155)
            fullcolor : Union[tuple[int, int, int], int, str], (optional)
                color of the bar when full
                defaults to (155, 70, 70)
            length : int, (optional)
                length of the slider bar
                defaults to 100
            count : int, (optional)
                number of frames to pass while inactive to send value
                defaults to 30
        """
        self.has_error = False
        if not (min_val <= value < max_val or min_val < value <= max_val):
            warn(
                f"ERROR [slider {self._name}] : wrong values Initialization, slider was not created"
            )
            self.has_error = True
        if thickness <= 0:
            warn(
                f"ERROR [slider {self._name}] : bad thickness value, slider was not created"
            )
            self.has_error = True
        if radius < thickness:
            warn(
                f"ERROR [slider {self._name}] : bad radius value, slider was not created"
            )
            self.has_error = True
        if length < 6 * radius:
            warn(
                f"ERROR [slider {self._name}] : bad length value, slider was not created"
            )
            self.has_error = True

        self._renderer = renderer
        self._x = x
        self._y = y
        self._name = name
        self._min_val = min_val
        self._max_val = max_val
        self._value = value
        self._incr = incr
        self._radius = radius
        self._thickness = thickness
        self._is_hidden = False
        self._click_count = count
        self._click = 0
        self._prev_state = True

        if isinstance(color, tuple) and len(color) == 3:
            self._color = color
        elif isinstance(color, int):
            self._color = color, color, color
        elif isinstance(color, str):
            try:
                self._color = COLORS[color.lower()]
            except KeyError:
                close = difflib.get_close_matches(color.lower(),
                                                  COLORS.keys(),
                                                  n=1,
                                                  cutoff=.5)[0]
                warn(
                    f"ERROR [slider {self._name}] : {color} is not a valid color name, using closest match {close} instead"
                )
                self._color = COLORS[close]
        else:
            warn(
                f"ERROR [slider {self._name}] : wrong color parameter, slider was not created"
            )
            self.has_error = True

        if isinstance(fullcolor, tuple) and len(fullcolor) == 3:
            self._fullcolor = fullcolor
        elif isinstance(fullcolor, int):
            self._fullcolor = fullcolor, fullcolor, fullcolor
        elif isinstance(fullcolor, str):
            try:
                self._fullcolor = COLORS[fullcolor.lower()]
            except KeyError:
                close = difflib.get_close_matches(fullcolor,
                                                  COLORS.keys(),
                                                  n=1,
                                                  cutoff=.5)[0]
                warn(
                    f"ERROR [slider {self._name}] : {fullcolor} is not a valid color name, using closest match {close} instead"
                )
                self._fullcolor = COLORS[close]
        else:
            warn(
                f"ERROR [slider {self._name}] : wrong full color parameter, slider was not created"
            )
            self.has_error = True

        if shape not in (SQUARE, CIRCLE, CROSS, PLUS):
            warn(
                f"ERROR [slider {self._name}] : wrong shape parameter, slider was not created"
            )
            self.has_error = True
        self._shape = shape

        self._length = length
        self._pad = self.length / (self.max_val - self.min_val)
        x = self._x + int(self._pad *
                          (self.value - self._min_val)) - self._radius
        self.rect = x + self._radius, self._y

    @property
    def click_count(self) -> int:
        """
        gets click_count of the slider
        """
        return self._click_count

    @click_count.setter
    def click_count(self, click_count: int) -> None:
        """
        sets the click_count of the slider\\
        deprecated : do not use

        Parameters
        ----------
            click_count : int
                new click_count
        """
        warn(
            f"INFO [slider {self._name}] : attempting to modify click behavior"
        )
        if click_count < 0:
            warn(
                f"ERROR [slider {self._name}] : bad click_count, nothing changed"
            )
            return
        self._click_count = click_count

    def click(self) -> None:
        """
        increments click
        """
        self._click += 1

    def reinit_click(self):
        """
        puts click back to 0
        """
        self._click = 0

    def check_click(self) -> bool:
        """
        check if click greater that click_count
        """
        return self._click >= self._click_count

    def is_active(self) -> bool:
        """
        if the slider is active\\
        i.e. its value was modified in the last frames (count setting)
        """
        return not self.check_click()

    def get_new_value(self) -> Union[float, None]:
        """
        value of slider if slider was activated and then idle\\
        will be ``None`` most of the time
        """
        if self._prev_state and not self.is_active():
            self._prev_state = self.is_active()
            return self.value
        self._prev_state = self.is_active()
        return None

    def _redo_rect(self) -> None:
        """
        sets correct coordinates to cursor
        """
        x = self._x + int(self._pad *
                          (self.value - self._min_val)) - self._radius
        self.rect = x + self._radius, self._y

    def _redo_pad(self) -> None:
        """
        sets correct padding
        """
        self._pad = self.length / (self.max_val - self.min_val)

    @property
    def value(self) -> float:
        """
        gets current slider value
        """
        return self._value

    @value.setter
    def value(self, value: float) -> None:
        """
        sets the value of the slider\\
        does not affect value if value is out of bounds

        Parameters
        ----------
            value : float
                new value
        """
        if not (self._min_val <= value < self._max_val
                or self._min_val < value <= self._max_val):
            warn(
                f"WARNING [slider {self._name}] : wrong value affectation, nothing happened"
            )
            return
        self._value = value
        self._redo_rect()

    @property
    def min_val(self) -> float:
        """
        gets current minimum slider value
        """
        return self._min_val

    @min_val.setter
    def min_val(self, min_val: float) -> None:
        """
        sets the minimum value of the slider\\
        does nothing if the minimum value goes above the maximum value or the current value\\
        deprecated : do not use

        Parameters
        ----------
            min_val : float
                new minimum value
        """
        warn(
            f"WARNING [slider {self._name}] : minimum value changing from {self._min_val} to {min_val}"
        )
        if not (min_val <= self._value < self._max_val
                or min_val < self._value <= self._max_val):
            warn(
                f"WARNING [slider {self._name}] : wrong minimum value affectation, nothing happened"
            )
            return
        self._min_val = min_val
        self._redo_pad()
        self._redo_rect()

    @property
    def max_val(self) -> float:
        """
        gets current maximum slider value
        """
        return self._max_val

    @max_val.setter
    def max_val(self, max_val: float) -> None:
        """
        sets the maximum value of the slider\\
        does nothing if the maximum value goes below the minimum value or the current value\\
        deprecated : do not use

        Parameters
        ----------
            max_val : float
                new maximum value
        """
        warn(
            f"WARNING [slider {self._name}] : maximum value changing from {self._max_val} to {max_val}"
        )
        if not (self._min_val <= self._value < max_val
                or self._min_val < self._value <= max_val):
            warn(
                f"WARNING [slider {self._name}] : wrong maximum value affectation, nothing happened"
            )
            return
        self._max_val = max_val
        self._redo_pad()
        self._redo_rect()

    @property
    def is_hidden(self) -> bool:
        """
        gets current display state of the slider
        """
        return self._is_hidden

    @is_hidden.setter
    def is_hidden(self, value: bool) -> None:
        """
        forces display state of the slider
        """
        self._is_hidden = value

    def hide(self) -> None:
        """
        hides the slider (does not display it automatically on the screen)\\
        slider value setting becomes unaccessible\\
        opposite method is ``reveal``
        """
        if self._is_hidden:
            warn(
                f"WARNING [slider {self._name}] : slider is already hidden, nothing changed"
            )
            return
        self._is_hidden = True

    def reveal(self) -> None:
        """
        reveals the slider back (displays it on the screen)\\
        slider value setting becomes accessible again\\
        opposite method is ``hide``
        """
        if not self._is_hidden:
            warn(
                f"WARNING [slider {self._name}] : slider is not hidden, nothing changed"
            )
            return
        self._is_hidden = False

    def move_to(self, x: int = None, y: int = None) -> None:
        """
        moves the slider to a designated location

        Parameters
        ----------
            x : int, (optional)
                x-coordinate
                defaults to None
            y : int, (optional)
                y-coordinate
                defaults to None
        """
        if x is not None:
            self._x = x
        if y is not None:
            self._y = y
        self._redo_rect()

    def resize(self, radius: int) -> None:
        """
        resize the sprite image

        Parameters
        ----------
            radius : int
                new radius
        """
        self._radius = radius
        self._redo_pad()
        self._redo_rect()

    @property
    def name(self) -> str:
        """
        gets the name of the slider
        """
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        """
        sets the name of the slider\\
        deprecated : do not use

        Parameters
        ----------
            name : str
                new name
        """
        warn(f"INFO [slider {self._name}] : name changing to {name}")
        self._name = name

    @property
    def thickness(self) -> int:
        """
        gets current slider thickness
        """
        return self._thickness

    @thickness.setter
    def thickness(self, thickness: int):
        """
        sets the thickness of the slider\\
        does nothing if thickness is 0 or less\\
        deprecated : do not use

        Parameters
        ----------
            thickness : int
                new thickness
        """
        warn(
            f"INFO [slider {self._name}] : thickness changing from {self._thickness} to {thickness}"
        )
        if thickness <= 0:
            warn(
                f"ERROR [slider {self._name}] : bad thickness value, nothing changed"
            )
            return
        self._thickness = thickness

    @property
    def color(self) -> tuple:
        """
        gets current slider line color
        """
        return self._color

    @color.setter
    def color(self, color: Union[tuple[int, int, int], int, str]) -> None:
        """
        sets the slider line color\\
        does not throw error if color is not matched\\
        deprecated : do not use

        Parameters
        ----------
            color : tuple | int | str
                the new color
        """
        warn(
            f"INFO [slider {self._name}] : color changing from {self._color} to {color}"
        )
        if isinstance(color, tuple) and len(color) == 3:
            self._color = color
        elif isinstance(color, int):
            self._color = color, color, color
        elif isinstance(color, str):
            try:
                self._color = COLORS[color.lower()]
            except KeyError:
                close = difflib.get_close_matches(color.lower(),
                                                  COLORS.keys(),
                                                  n=1,
                                                  cutoff=.5)[0]
                warn(
                    f"ERROR [slider {self._name}] : {color} is not a valid color name, using closest match {close} instead"
                )
                self._color = COLORS[close]
        else:
            warn(
                f"ERROR [slider {self._name}] : wrong color parameter, nothing changed"
            )

    @property
    def fullcolor(self) -> tuple:
        """
        gets current slider full line color
        """
        return self._fullcolor

    @fullcolor.setter
    def fullcolor(self, fullcolor: Union[tuple[int, int, int], int,
                                         str]) -> None:
        """
        sets the slider full line color\\
        does not throw error if color is not matched\\
        deprecated : do not use

        Parameters
        ----------
            fullcolor : tuple | int | str
                the new color
        """
        warn(
            f"INFO : [slider {self._name}] color changing from {self._fullcolor} to {fullcolor}"
        )
        if isinstance(fullcolor, tuple) and len(fullcolor) == 3:
            self._fullcolor = fullcolor
        elif isinstance(fullcolor, int):
            self._fullcolor = fullcolor, fullcolor, fullcolor
        elif isinstance(fullcolor, str):
            try:
                self._fullcolor = COLORS[fullcolor.lower()]
            except KeyError:
                close = difflib.get_close_matches(fullcolor,
                                                  COLORS.keys(),
                                                  n=1,
                                                  cutoff=.5)[0]
                warn(
                    f"ERROR [slider {self._name}] : {fullcolor} is not a valid color name, using closest match {close} instead"
                )
                self._fullcolor = COLORS[close]
        else:
            warn(
                f"ERROR [slider {self._name}] : wrong color parameter, nothing changed"
            )

    @property
    def radius(self) -> int:
        """
        gets slider radius
        """
        return self._radius

    @radius.setter
    def radius(self, radius: int) -> None:
        """
        sets the slider cursor radius\\
        does nothing if radius less than thickness\\
        deprecated : do not use

        Parameters
        ----------
            radius : int
                new radius
        """
        warn(
            f"INFO [slider {self._name}] : radius changing from {self._radius} to {radius}"
        )
        if radius < self._thickness:
            warn(
                f"ERROR [slider {self._name}] : bad radius value, nothing changed"
            )
            return
        self._radius = radius
        self._redo_rect()

    @property
    def shape(self) -> str:
        """
        gets current shape of the slider cursor
        """
        return self._shape

    @shape.setter
    def shape(self, shape: str) -> None:
        """
        sets current shape of the slider\\
        deprecated : do not use

        Parameters
        ----------
            shape : str
                new shape
                SQUARE | CIRCLE | CROSS | PLUS
        """
        warn(f"INFO [slider {self._name}] : attempting shape change")
        if shape not in (SQUARE, CIRCLE, CROSS, PLUS):
            warn(
                f"ERROR [slider {self._name}] : {shape} is not a valid shape, nothing changed"
            )
            return
        self._shape = shape

    @property
    def length(self) -> int:
        """
        gets slider length
        """
        return self._length

    @length.setter
    def length(self, length: int) -> None:
        """
        sets the slider bar length\\
        does nothing is length is less than 3 times the diameter\\
        deprecated : do not use

        Parameters
        ----------
            length : int
                new length
        """
        warn(
            f"INFO [slider {self._name}] : length changing from {self._length} to {length}"
        )
        if length < 6 * self._radius:
            warn(f"ERROR [slider {self._name}] : bad length value")
            return
        self._length = length
        self._redo_pad()
        self._redo_rect()

    def set_value(self, pos: tuple) -> None:
        """
        sets slider value based on mouse position

        Parameters
        ----------
            pos : tuple
                mouse position
        """
        x = pos[0]
        if 0 <= (x_rel := x - self._x) <= self.length:
            val = _map(x_rel, 0, self.length, self.min_val, self.max_val)
            self.value = round(val, self._incr)

    def collide(self, pos: tuple[int, int]) -> bool:
        """
        collision check

        Parameters
        ----------
            pos : tuple
                mouse position
        """
        x, y = pos
        return self._x <= x <= self._x + self.length and self._y - self.radius <= y <= self._y + self.radius

    def draw(self) -> None:
        """
        draws the slider on the screen\\
        draws line and apply text
        """
        renderer = self._renderer
        renderer.push()

        name_label = renderer.font.render(self.name, True, (0, 0, 0))
        min_label = renderer.font.render(str(round(self.min_val, self._incr)),
                                         True, (0, 0, 0))
        max_label = renderer.font.render(str(round(self.max_val, self._incr)),
                                         True, (0, 0, 0))
        val_label = renderer.font.render(str(self.value), True, (0, 0, 0))

        # renderer.stroke_weight = self.thickness
        # renderer.stroke = self.color
        pygame.draw.line(renderer._window, self.color, (self._x, self._y),
                         (self._x + self.length, self._y), self.thickness)
        # renderer.line((self._x, self._y), (self._x + self.length, self._y))

        # renderer.stroke = self.fullcolor
        pad = self.length / (self.max_val - self.min_val)
        x = self._x + int(pad * (self.value - self._min_val))
        pygame.draw.line(renderer._window, self.fullcolor, (self._x, self._y),
                         (x, self._y), self.thickness)
        # renderer.line((self._x, self._y), (x, self._y))

        if self.shape == SQUARE:
            # renderer.fill = self.fullcolor
            # renderer.no_stroke()
            # renderer.rect_mode = CENTER
            rect = self.rect[0] - self.radius, self.rect[
                1] - self.radius, 2 * self.radius, 2 * self.radius
            pygame.draw.rect(renderer._window, self.fullcolor, rect, 0)
            # renderer.rect(self.rect, 2 * self.radius, 2 * self.radius)
        elif self.shape == CIRCLE:
            # renderer.fill = self.fullcolor
            # renderer.no_stroke()
            # renderer.rect_mode = CENTER
            pygame.draw.circle(renderer._window, self.fullcolor, self.rect,
                               self.radius, 0)
            # renderer.circle(self.rect, self.radius)
        elif self.shape == CROSS:
            # renderer.no_fill()
            # renderer.stroke = self.fullcolor
            # renderer.stroke_weight = self.thickness
            pygame.draw.line(
                renderer._window, self.fullcolor,
                (self.rect[0] - self.radius, self.rect[1] - self.radius),
                (self.rect[0] + self.radius, self.rect[1] + self.radius),
                self.thickness)
            pygame.draw.line(
                renderer._window, self.fullcolor,
                (self.rect[0] - self.radius, self.rect[1] + self.radius),
                (self.rect[0] + self.radius, self.rect[1] - self.radius),
                self.thickness)
            # renderer.line((self.rect[0] - self.radius, self.rect[1] - self.radius),
            #               (self.rect[0] + self.radius, self.rect[1] + self.radius))
            # renderer.line((self.rect[0] - self.radius, self.rect[1] + self.radius),
            #               (self.rect[0] + self.radius, self.rect[1] - self.radius))
        elif self.shape == PLUS:
            # renderer.no_fill()
            # renderer.stroke = self.fullcolor
            # renderer.stroke_weight = self.thickness
            pygame.draw.line(renderer._window, self.fullcolor,
                             (self.rect[0], self.rect[1] + self.radius),
                             (self.rect[0], self.rect[1] - self.radius),
                             self.thickness)
            pygame.draw.line(renderer._window, self.fullcolor,
                             (self.rect[0] - self.radius, self.rect[1]),
                             (self.rect[0] + self.radius, self.rect[1]),
                             self.thickness)
            # renderer.line((self.rect[0], self.rect[1] + self.radius),
            #               (self.rect[0], self.rect[1] - self.radius))
            # renderer.line((self.rect[0] - self.radius, self.rect[1]),
            #               (self.rect[0] + self.radius, self.rect[1]))

        renderer.text(self._x - name_label.get_width() - 10,
                      self._y - name_label.get_height() // 2, self.name)
        renderer.text(self._x - min_label.get_width() // 2,
                      self._y + self.thickness,
                      str(round(self.min_val, self._incr)))
        renderer.text(self._x + self.length - max_label.get_width() // 2,
                      self._y + self.thickness,
                      str(round(self.max_val, self._incr)))
        renderer.text(self.rect[0] - val_label.get_width() // 2,
                      self.rect[1] - self.radius - val_label.get_height(),
                      str(self.value))

        renderer.pop()



from typing import Union
from ..data import *

__all__ = ["Button"]

import difflib
import pygame


class Button:
    """
    Pygame Button
    =============
    created by ``Renderer``
    """
    def __init__(self,
                 renderer,
                 x: int,
                 y: int,
                 name: str,
                 count: int = 1,
                 action=lambda: None,
                 width: int = 50,
                 height: int = 50,
                 shape: str = RECTANGLE,
                 color: Union[tuple[int, int, int], int, str] = None,
                 stroke: Union[tuple[int, int, int], int, str] = None,
                 weight: int = 1) -> None:
        """
        new Button instance

        Parameters
        ----------
            renderer : Renderer
                the Renderer instance the Button is linked to
            x : int
                the x-coordinate (TOP LEFT of the button)
            y : int
                the y-coordinate (TOP LEFT of the button)
            name : str
                the name of the button (must be unique !)
            count : int, (optional)
                number of frames to pass while un-clicked to be able to trigger the button again
                defaults to 1
            action : python function, (optional)
                function to trigger when pressed
                defaults to lambda:None
            width : int, (optional)
                the width of the button
                defaults to 50
            height : int, (optional)
                the height of the button
                defaults to 50
            shape : str, (optional)
                the shape of the button
                RECTANGLE | ELLIPSE
                defaults to RECTANGLE
            color : Union[tuple[int, int, int], int, str], (optional)
                color to fill button and enables filling
                if both color and stroke are None, the button will be filled by default
                defaults to None
            stroke : Union[tuple[int, int, int], int, str], (optional)
                color to draw the outside box and enables stroking
                defaults to None
            weight : int, (optional)
                stroke weight if stroke is not None
                defaults to 1
        """
        self.has_error = False
        self._renderer = renderer
        self._x = x
        self._y = y
        self._name = name
        self._click_count = count
        self._click = 0
        self._action = action
        self._width = width
        self._height = height
        self._is_hidden = False

        color = (color, 0)[color is None and stroke is None]
        if isinstance(color, tuple) and len(color) == 3:
            self._color = color
        elif isinstance(color, int):
            self._color = color, color, color
        elif isinstance(color, str):
            try:
                self._color = COLORS[color.lower()]
            except KeyError:
                close = difflib.get_close_matches(color.lower(),
                                                  COLORS.keys(),
                                                  n=1,
                                                  cutoff=.5)[0]
                warn(
                    f"ERROR [button {self._name}] : {color} is not a valid color name, using closest match {close} instead"
                )
                self._color = COLORS[close]
        elif color is None:
            self._color = None
        else:
            warn(
                f"ERROR [button {self._name}] : wrong color parameter, button was not created"
            )
            self.has_error = True

        if isinstance(stroke, tuple) and len(stroke) == 3:
            self._stroke = stroke
        elif isinstance(stroke, int):
            self._stroke = stroke, stroke, stroke
        elif isinstance(stroke, str):
            try:
                self._stroke = COLORS[stroke.lower()]
            except KeyError:
                close = difflib.get_close_matches(stroke,
                                                  COLORS.keys(),
                                                  n=1,
                                                  cutoff=.5)[0]
                warn(
                    f"ERROR [button {self._name}] : {stroke} is not a valid color name, using closest match {close} instead"
                )
                self._stroke = COLORS[close]
        elif stroke is None:
            self._stroke = None
        else:
            warn(
                f"ERROR [button {self._name}] : wrong stroke parameter, button was not created"
            )
            self.has_error = True

        if shape not in (RECTANGLE, ELLIPSE):
            warn(
                f"ERROR [button {self._name}] : wrong shape parameter, button was not created"
            )
            self.has_error = True
        self._shape = shape

        if weight <= 0 and self._stroke is not None:
            warn(
                f"ERROR [button {self._name}] : weight can't be {weight} if stroking is not disabled, button was not created"
            )
            self.has_error = True
        self._weight = weight

    @property
    def click_count(self) -> int:
        """
        gets click_count of the button
        """
        return self._click_count

    @click_count.setter
    def click_count(self, click_count: int) -> None:
        """
        sets the click_count of the button\\
        deprecated : do not use

        Parameters
        ----------
            click_count : int
                new click_count
        """
        warn(
            f"INFO [button {self._name}] : attempting to modify click behavior"
        )
        if click_count < 0:
            warn(
                f"ERROR [button {self._name}] : bad click_count, nothing changed"
            )
            return
        self._click_count = click_count

    def click(self) -> None:
        """
        increments click
        """
        self._click += 1

    def reinit_click(self):
        """
        puts click back to 0
        """
        self._click = 0

    def check_click(self) -> bool:
        """
        check if click greater that click_count
        """
        return self._click >= self._click_count

    def on_press(self):
        """
        triggers the button action and returns the result
        """
        trigger = self._action()
        return trigger

    @property
    def is_hidden(self) -> bool:
        """
        gets current display state of the button
        """
        return self._is_hidden

    @is_hidden.setter
    def is_hidden(self, value: bool) -> None:
        """
        forces the display state of the button
        """
        self._is_hidden = value

    @property
    def shape(self) -> str:
        """
        gets current shape to draw the button
        """
        return self._shape

    @shape.setter
    def shape(self, shape: str) -> None:
        """
        sets the shape to draw the button\\
        deprecated : do not use

        Parameters
        ----------
            shape : str
                the shape of the button
                RECTANGLE | ELLIPSE
        """
        warn(f"INFO [button {self._name}] : attempting shape change")
        if shape not in (RECTANGLE, ELLIPSE):
            warn(
                f"ERROR [button {self._name}] : {shape} is not a valid shape, nothing changed"
            )
            return
        self._shape = shape

    @property
    def color(self) -> tuple[int, int, int]:
        """
        gets button filling color\\
        might be None if filling is disabled for this button
        """
        return self._color

    @color.setter
    def color(self, color: Union[tuple[int, int, int], int, str]) -> None:
        """
        sets button filling color\\
        color can be None to disable filling\\
        deprecated : do not use

        Parameters
        ----------
            color : tuple | int | str
                the new color
        """
        warn(f"INFO [button {self._name}] : attempting filling change")
        if isinstance(color, tuple) and len(color) == 3:
            self._color = color
        elif isinstance(color, int):
            self._color = color, color, color
        elif isinstance(color, str):
            try:
                self._color = COLORS[color.lower()]
            except KeyError:
                close = difflib.get_close_matches(color.lower(),
                                                  COLORS.keys(),
                                                  n=1,
                                                  cutoff=.5)[0]
                warn(
                    f"ERROR [button {self._name}] : {color} is not a valid color name, using closest match {close} instead"
                )
                self._color = COLORS[close]
        elif color is None:
            if self._stroke is None:
                warn(
                    f"ERROR [button {self._name}] : filling can't be disabled if stroking also is, nothing changed"
                )
                return
            self._color = None
        else:
            warn(
                f"ERROR [button {self._name}] : {color} is not a valid color, nothing changed"
            )

    @property
    def stroke(self) -> tuple[int, int, int]:
        """
        gets button stroking color\\
        might be None if stroking is disabled for this button
        """
        return self._stroke

    @stroke.setter
    def stroke(self, stroke: Union[tuple[int, int, int], int, str]) -> None:
        """
        sets button stroking color\\
        color can be None to disable stroking\\
        deprecated : do not use

        Parameters
        ----------
            stroke : None | tuple | int | str
                the new color
        """
        if isinstance(stroke, tuple) and len(stroke) == 3:
            self._stroke = stroke
        elif isinstance(stroke, int):
            self._stroke = stroke, stroke, stroke
        elif isinstance(stroke, str):
            try:
                self._stroke = COLORS[stroke.lower()]
            except KeyError:
                close = difflib.get_close_matches(stroke,
                                                  COLORS.keys(),
                                                  n=1,
                                                  cutoff=.5)[0]
                warn(
                    f"ERROR [button {self._name}] : {stroke} is not a valid color name, using closest match {close} instead"
                )
                self._stroke = COLORS[close]
        elif stroke is None:
            if self._color is None:
                warn(
                    f"ERROR [button {self._name}] : stroking can't be disabled if filling also is, nothing changed"
                )
                return
            self._stroke = None
        else:
            warn(
                f"ERROR [button {self._name}] : {stroke} is not a valid color, nothing changed"
            )

    @property
    def weight(self) -> int:
        """
        gets stroke weight for this button
        """
        return self._weight

    @weight.setter
    def weight(self, weight: int) -> None:
        """
        sets stroke weight for this button\\
        can be any integer if stroking is disabled

        Parameters
        ----------
            weight : int
                the new weight
        """
        if weight <= 0 and self._stroke is not None:
            warn(
                f"ERROR [button {self._name}] : weight can't be {weight} if stroking is not disabled, nothing changed"
            )
            return
        self._weight = weight

    def hide(self) -> None:
        """
        hides the button (does not display it automatically on the screen)\\
        button action becomes unaccessible\\
        opposite method is ``reveal``
        """
        if self._is_hidden:
            warn(
                f"WARNING [button {self._name}] : button is already hidden, nothing changed"
            )
            return
        self._is_hidden = True

    def reveal(self) -> None:
        """
        reveals the button back (displays it on the screen)\\
        button action becomes accessible again\\
        opposite method is ``hide``
        """
        if not self._is_hidden:
            warn(
                f"WARNING [button {self._name}] : button is not hidden, nothing changed"
            )
            return
        self._is_hidden = False

    def move_to(self, x: int = None, y: int = None) -> None:
        """
        moves the button to a designated location

        Parameters
        ----------
            x : int, (optional)
                x-coordinate
                defaults to None
            y : int, (optional)
                y-coordinate
                defaults to None
        """
        if x is not None:
            self._x = x
        if y is not None:
            self._y = y

    def resize(self, width: int, height: int) -> None:
        """
        resize the button box

        Parameters
        ----------
            width : int
                new width
            height : int
                new height
        """
        if height <= 2 and width <= 2:
            warn(
                f"ERROR [button {self._name}] : ({width}, {height}) is not a valid size, nothing changed"
            )
            return
        self._height = height
        self._width = width

    @property
    def name(self) -> str:
        """
        gets the name of the button
        """
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        """
        sets the name of the button\\
        deprecated : do not use

        Parameters
        ----------
            name : str
                new name
        """
        warn(f"INFO [button {self._name}] : name changing to {name}")
        self._name = name

    @property
    def action(self):
        """
        gets the function of the button
        """
        return self._action

    @action.setter
    def action(self, action) -> None:
        """
        sets the action of the button\\
        deprecated : do not use

        Parameters
        ----------
            action : python function
                new action
        """
        warn(f"INFO [button {self._name}] : action changed")
        self._action = action

    def collide(self, pos: tuple[int, int]) -> bool:
        """
        collision check
        """
        x, y = pos
        return self._x <= x <= self._x + self._width and self._y <= y <= self._y + self._height

    def draw(self) -> None:
        """
        draws the button on the screen\\
        draws box and apply text
        """
        renderer = self._renderer

        renderer.push()
        renderer.no_fill()
        renderer.no_stroke()

        name_label = renderer.font.render(self.name, True, (0, 0, 0))

        x = self._x + (self._width // 2)
        y = self._y + (self._height // 2)
        # renderer.rect_mode = CENTER

        # if self.color is not None:
        #     renderer.fill = self.color
        # if self.stroke is not None:
        #     renderer.stroke = self.stroke
        #     renderer.stroke_weight = self.weight

        if self.shape == RECTANGLE:
            if self.color is not None:
                pygame.draw.rect(renderer._window, self.color,
                                 (self._x, self._y, self._width, self._height),
                                 0)
            if self.stroke is not None:
                pygame.draw.rect(renderer._window, self.stroke,
                                 (self._x, self._y, self._width, self._height),
                                 self.weight)
            # renderer.rect((x, y), self._width, self._height)
        elif self.shape == ELLIPSE:
            if self.color is not None:
                pygame.draw.ellipse(
                    renderer._window, self.color,
                    (self._x, self._y, self._width, self._height), 0)
            if self.stroke is not None:
                pygame.draw.ellipse(
                    renderer._window, self.stroke,
                    (self._x, self._y, self._width, self._height), self.weight)
            # renderer.ellipse((x, y), self._width, self._height)

        x -= name_label.get_width() // 2
        y -= name_label.get_height() // 2
        renderer.text(x, y, self.name)

        renderer.pop()



"""
Pygame Engine
=============

Provides
 1. Vector support for ``python 3.9`` and upper in two or three dimensional space
 2. Fast standards operations using numpy
 3. Full random support plus ndarray and p5 compatibility
 4. 2D visual renderer using ``pygame``
 5. Fast drawing features and global settings
 6. Fast 2D, 3D, and 4D Noise algorithms (Open Simplex and Perlin) for smooth randomness
 7. A physics sandbox for basic simulation

Initialization
--------------
>>> # from now on we will assume Renderer and Vector are imported as followed
... from phoenyx import *
>>> # new pygame Renderer
... renderer = Renderer(600, 600)

Drawing basics
--------------
>>> renderer.stroke_weight = 2
... # makes lines appear thicker
>>> renderer.stroke = 255
... # will draw white lines
>>> renderer.circle((300, 300), 100)
... # will draw a white circle with a width of 2 at 300 300 with a radius of 100

Buttons
-------
You can creates Buttons with
>>> renderer.create_button(500, 500, "test", action=lambda: print("button pressed"))
... # creates a new button at 500 500 named test
... # which prints "button pressed" when pressed

Generally speaking the action of the button is performed each time the button is
pressed but the result is not accessible (best is to throw functions). Note that
you can customize the number of frames each button has to wait while unpressed to
be able to be triggered again.

Sliders
-------
You can also create sliders with
>>> renderer.create_slider(100, 100, "slider", 0, 10, 5, 0, length=200)
... # creates a new slider at 100 100 named "slider"
... # with a minimum value of 0, a maximum value of 10
... # starts at the value 5
... # have a floating point precision of 0 decimals
... # with an additional argument which makes its length 200

All sliders can return their value based on their name (which should be unique)
and the update of their value is done automatically. You must however take their
value and then use it manually (it is not bound to an external variable). Note
that each slider greatly decreases the frames of the ``Renderer``.

Menus
-----
You now can create side menus with
>>> renderer.create_menu("menu", test1=lambda: print("test1"), test2=lambda: print("test2"))
... # creates a new menu on the right of the screen
... # which has 2 buttons when expanded
... # the first one printing "test1" in the console
... # and the second printing "test2"

It is worth noting that you can only create 2 side menus, the first one being
on the right of the screen (default side) and the second being on the left. Also
note that extensive actions list might not show up properly depending on the
size of the window. The menu background will show up on the top of all other
drawings and will be the same color as the window background unless otherwise
specified.

Vectors
-------
A Vector -- specifically an Euclidean (or geometric) vector -- in
two or three dimensional space is a geometric entity that has some
magnitude (or length) and a direction.

>>> zero = Vector()
>>> zero
Vector(0.00, 0.00, 0.00)

>>> vec_2d = Vector(3, 4)
>>> vec_2d
Vector(3.00, 4.00, 0.00)

>>> vec_3d = Vector(2, 3, 4)
>>> vec_3d
Vector(2.00, 3.00, 4.00)

Open Simplex Noise
------------------
OpenSimplex n-dimensional gradient noise functions. Support for 2 to 4
dimensional evaluation and integrated to Phoenyx.

Based on a modified Simplex Noise algorithm, Open Simplex Noise by Curt Spencer.
Simplex Noise is a beautification of the Perlin Noise algorithm.

Initialization is using a permutation array generated from a 64-bit seed number.

Perlin Noise
------------
Perlin Noise n-dimensional gradient noise functions. Integrated to Phoenyx

Callable that produces Perlin noise for an arbitrary point in an
arbitrary number of dimensions.  The underlying grid is aligned with the
integers.

There is no limit to the coordinates used, new gradients are generated on
the fly as necessary.

More octaves create a foggier and more-detailed noise pattern.  More
than 4 octaves is rather excessive.

``tile`` can be used to make a seamlessly tiling pattern.  For example:
>>> noise = PerlinNoise(2, tile=(0, 3))
This will produce noise that tiles every 3 units vertically, but never
tiles horizontally.

If ``unbias`` is true, the quintic function will be applied to the
output before returning it, to counteract some of Perlin noise's
significant bias towards the center of its output range.

SandBox
-------
Since v0.2.0 you can create a basic physics engine. It handles the creation of new bodies, some collisions and bouncing on the edges of the world boundaries. It also gives all bodies a default drawing method but you should create your own by inheriting the Body class and modifying what you want. Note that all bodies leaving the world are lost (that applies to bodies that do not teleport around the edges or bounce on the edges of the world). Also note that the SandBox has its center be the center of the renderer window.

# ``SandBox`` object that has a world the same dimensions as the ``Renderer`` window
>>> sandbox = SandBox(renderer, renderer.win_width/2, renderer.win_height/2, bounce=True)

All bodies will bounce on the edges of the screen (bouncing is based on the center of the bodies).

Please note
-----------
Please note that this library is not fully tested and thus may be very buggy.
So pay attention especially when creating buttons / sliders and attempting to
remove or trigger them.
Also both buttons and sliders currently have unsupported methods for the Renderer
such as moving them on the screen, resizing them, changing their attributes...
Finally, ERROR and WARNING do not cause a 'real' ``python error`` but throw some
pieces of information in the console.
"""

from .core import *
from .data import *
from .pmath import *
from .elements import *

# Hello from Phoenyx
print(f"Hello from Phoenyx - please visit us on GitHub")

from .__version__ import __title__
from .__version__ import __description__
from .__version__ import __url__
from .__version__ import __version__
from .__version__ import __author__
from .__version__ import __license__



import os
import setuptools

here = os.path.abspath(os.path.dirname(__file__))

packages = ["phoenyx"]
meta_data: dict[str, str] = {}

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

with open("requirements.txt", "r", encoding="utf-8") as f:
    requires = f.read().split("\n")

with open(os.path.join(here, "phoenyx", "__version__.py"),
          "r",
          encoding="utf-8") as f:
    exec(f.read(), meta_data)

setuptools.setup(
    name=meta_data["__title__"],
    version=meta_data["__version__"],
    author=meta_data["__author__"],
    author_email=meta_data["__author_email__"],
    description=meta_data["__description__"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=meta_data["__url__"],
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Topic :: Education",
        "Topic :: Multimedia :: Graphics",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9",
    install_requires=requires,
)



from typing import Callable, Union
import pygame
import difflib
import math as m

pygame.init()

__all__ = ["Renderer"]

import __main__  # type: ignore (pylance bad)

from ..data import *
from ..pmath import *
from ..elements import *


class Renderer:
    """
    Phoenyx Renderer
    ================
    Provides:
    1. 2D visual renderer using ``pygame`` on ``python 3.9`` and above
    2. fast drawing features and global settings
    3. plenty other classes to ease drawing and focus on programming

    Please go through exemples, in-file docstrings and methods, and tests files.

    Initialization
    --------------
    >>> # from now on we will assume Renderer is imported as followed
    ... from renderer import Renderer
    >>> renderer = Renderer(600, 600)

    Drawing basics
    --------------
    >>> renderer.stroke_weight = 2
    ... # makes lines appear thicker
    >>> renderer.stroke = 255
    ... # will draw white lines
    >>> renderer.circle((300, 300), 100)
    ... # will draw a white circle with a width of 2 at 300 300 with a radius of 100

    Buttons
    -------
    You can creates Buttons with
    >>> renderer.create_button(500, 500, "test", action=lambda: print("button pressed"))
    ... # creates a new button at 500 500 named test
    ... # which prints "button pressed" when pressed

    Generally speaking the action of the button is performed each time the button is
    pressed but the result is not accessible (best is to throw functions). Note that
    you can customize the number of frames each button has to wait while unpressed to
    be able to be triggered again.

    Sliders
    -------
    You can also create sliders with
    >>> renderer.create_slider(100, 100, "slider", 0, 10, 5, 0, length=200)
    ... # creates a new slider at 100 100 named "slider"
    ... # with a minimum value of 0, a maximum value of 10
    ... # starts at the value 5
    ... # have a floating point precision of 0 decimals
    ... # with an additional argument which makes its length 200

    All sliders can returns their value based on their name (which should be unique)
    and the update of their value is done automatically. You must however take their
    value and then use it manually (it is not bound to an external variable). Note
    that each slider greatly decreases the frames of the ``Renderer``.

    Menus
    -----
    You now can create side menus with
    >>> renderer.create_menu("menu", test1=lambda: print("test1"), test2=lambda: print("test2"))
    ... # creates a new menu on the right of the screen
    ... # which has 2 buttons when expanded
    ... # the first one printing "test1" in the console
    ... # and the second printing "test2"

    It is worth noting that you can only create 2 side menus, the first one being
    on the right of the screen (default side) and the second being on the left. Also
    note that extensive actions list might not show up properly depending on the
    size of the window. The menu background will show up on the top of all other
    drawings and will be the same color as the window background unless otherwise
    specified.

    Please note
    -----------
    Please note that this library is not fully tested and thus may be very buggy.
    So pay attention especially when creating buttons / sliders and attempting to
    remove or trigger them.
    Also both buttons and sliders currently have unsupported methods for the Renderer
    such as moving them on the screen, resizing them, changing their attributes...
    Finally, ERROR and WARNING do not cause a 'real' ``python error`` but throw some
    pieces of information in the console.
    """

    # font = pygame.font.SysFont("comicsans", 11)

    def __init__(self, width: int, height: int, title: str = None) -> None:
        """
        new Renderer instance

        Parameters
        ----------
            width : int
                width of the window
            height : int
                height of the window
            title : str, (optional)
                title of the window, can be changed
                defaults to None
        """
        # window management
        self._window: pygame.Surface = pygame.display.set_mode((width, height))
        self._width = width
        self._height = height
        self._title = (title, "Pygame Engine with Python")[title is None]
        pygame.display.set_caption(self._title)

        self.font = pygame.font.SysFont("comicsans", 11)
        self._ff_name = "comicsans"
        self._ff_is_sys = True

        # drawing attributes management
        self._fill_color = (255, 255, 255)
        self._fill = True
        self._stroke_color = (255, 255, 255)
        self._stroke_weight = 1
        self._stroke = True
        self._rect_mode = CORNER

        # offsets
        self._has_translation = False
        self._x_offset = 0
        self._y_offset = 0
        self._translation_behavior = RESET

        # angles
        self._has_rotation = False
        self._rot_angle = 0
        self._rotation_behavior = RESET

        # scale
        self._has_scale = False
        self._scale_factor = 1
        self._scale_behavior = RESET

        # background
        self._has_auto_bg = False
        self._bg = (51, 51, 51)

        # shapes
        self._is_drawing_shape = False
        self._all_vertexes: list[tuple] = []

        # text management
        self._text_color = (255, 255, 255)
        self._text_size = 12

        # buttons management
        self._has_buttons = False
        self._all_buttons: set[Button] = set()

        # sliders management
        self._has_sliders = False
        self._all_sliders: set[Slider] = set()

        # menus management
        self._has_left_menu = False
        self._has_right_menu = False
        self._all_menus: set[Menu] = set()

        # scrollbar management
        self._has_scrollbar = False
        self._scrollbar: ScrollBar = None
        self._ps_value = 0

        # pixels
        self._pixels: pygame.PixelArray = None
        self._is_p_loaded = False

        # fps
        self._fps = 60
        self._clock = pygame.time.Clock()

        # save
        self._has_save = False
        # list[list[tuple[int, int, int], bool, tuple[int, int, int], int, bool, bool, int, int, bool,
        #           Union[int, float], bool, Union[int, float], str, str, str, str, tuple[int, int, int], int]]
        self._save: list[list[18]] = []

        # keys
        self._key_binding: dict[int, int] = {}
        self._actions: list[Callable[[], None]] = []
        self._keys_behavior: list[str] = []
        self._pressed: dict[int, bool] = {}
        self.keys = Keys()

        # bench mode
        self._benchmark = False

        # running
        self._is_running = True

    def set_title(self, title: str) -> None:
        """
        gives a new title to the main window

        Parameters
        ----------
            title : str
                the new title
        """
        self._title = title
        pygame.display.set_caption(self._title)

    def set_icon(self, path: str) -> None:
        """
        sets the icon of the app
        """
        surface = pygame.image.load(path)
        pygame.display.set_icon(surface)

    @property
    def win_width(self) -> int:
        """
        gets current window width
        """
        return self._width

    @property
    def win_height(self) -> int:
        """
        gets current window height
        """
        return self._height

    @property
    def win_bg(self) -> tuple[int, int, int]:
        """
        gets current window filling background color\\
        might be wrong if the ``background`` method was never called
        """
        return self._bg

    @property
    def mouse_pos(self) -> tuple[int, int]:
        """
        gets current mouse position as a [int, int] tuple
        """
        return pygame.mouse.get_pos()

    @property
    def mouse_x(self) -> int:
        """
        gets current position of the mouse cursor along the x-axis
        """
        return self.mouse_pos()[0]

    @property
    def mouse_y(self) -> int:
        """
        gets current position of the mouse cursor along the y-axis
        """
        return self.mouse_pos()[1]

    def mouse_is_down(self, button: int = 0) -> bool:
        """
        gets current state of the mouse button

        Parameters
        ----------
            button : int, (optional)
                0 | 1 | 2 button of the mouse to check
                defaults to 0
        """
        return pygame.mouse.get_pressed(num_buttons=3)[button] != 0

    def no_fill(self) -> None:
        """
        disables filling globally
        """
        self._fill = False

    @property
    def fill(self) -> tuple[int, int, int]:
        """
        gets current fill color event if stroking is disabled

        Returns
        -------
            tuple : color
        """
        return self._fill_color

    @fill.setter
    def fill(self, color: Union[tuple[int, int, int], int, str]) -> None:
        """
        changes the fill color globally

        Parameters
        ----------
            color : tuple | int | str
                the new color
        """
        self._fill = True
        if isinstance(color, tuple) and len(color) == 3:
            self._fill_color = color
        elif isinstance(color, int):
            self._fill_color = color, color, color
        elif isinstance(color, str):
            try:
                self._fill_color = COLORS[color.lower()]
            except KeyError:
                close = difflib.get_close_matches(color.lower(),
                                                  COLORS.keys(),
                                                  n=1,
                                                  cutoff=.5)[0]
                warn(
                    f"ERROR [renderer] : {color} is not a valid color name, using closest match {close} instead"
                )
                self._fill_color = COLORS[close]
        else:
            warn(
                f"ERROR [renderer] : {color} not a valid color parameter, nothing changed"
            )

    def no_stroke(self) -> None:
        """
        disables stroking globally
        """
        self._stroke = False

    @property
    def stroke(self) -> tuple[int, int, int]:
        """
        gets current stroke color event if stroking is disabled

        Returns
        -------
            tuple : color
        """
        return self._stroke_color

    @stroke.setter
    def stroke(self, color: Union[tuple[int, int, int], int, str]) -> None:
        """
        changes the stroke color globally

        Parameters
        ----------
            color : tuple | int | str
                the new color
        """
        self._stroke = True
        if isinstance(color, tuple) and len(color) == 3:
            self._stroke_color = color
        elif isinstance(color, int):
            self._stroke_color = color, color, color
        elif isinstance(color, str):
            try:
                self._stroke_color = COLORS[color.lower()]
            except KeyError:
                close = difflib.get_close_matches(color.lower(),
                                                  COLORS.keys(),
                                                  n=1,
                                                  cutoff=.5)[0]
                warn(
                    f"ERROR [renderer] : {color} is not a valid color name, using closest match {close} instead"
                )
                self._stroke_color = COLORS[close]
        else:
            warn(
                f"ERROR [renderer] : {color} not a valid color parameter, nothing changed"
            )

    @property
    def stroke_weight(self) -> int:
        """
        gets current stroke weight event if stroking is disabled

        Returns
        -------
            int: stroke weight
        """
        return self._stroke_weight

    @stroke_weight.setter
    def stroke_weight(self, weight: int) -> None:
        """
        changes the stroke weight globally

        Parameters
        ----------
            weight : int
                new stroke weight
        """
        self._stroke = True
        self._stroke_weight = weight

    @property
    def rect_mode(self) -> str:
        """
        gets current rect mode
        """
        return self._rect_mode

    @rect_mode.setter
    def rect_mode(self, mode: str = CENTER) -> None:
        """
        changes rect mode globally\\
        does not enables stroking neither filling back

        Parameters
        ----------
            mode : str, (optional)
                CENTER or CORNER
                defaults to CENTER
        """
        if mode not in (CENTER, CORNER):
            warn(
                f"ERROR [renderer] : {mode} is not a valid mode for rect_mode, nothing happened"
            )
            return
        self._rect_mode = mode

    @property
    def text_size(self) -> int:
        """
        gets the current text size
        """
        return self._text_size

    @text_size.setter
    def text_size(self, size: int) -> None:
        """
        changes text size globally

        Parameters
        ----------
            size : int
                new text size
        """
        self._text_size = size
        if self._ff_is_sys:
            self.font = pygame.font.SysFont(self._ff_name, size)
        else:
            self.font = pygame.font.Font(self._ff_name, size)

    @property
    def text_font(self) -> str:
        """
        name or path of the font in use
        """
        return self._ff_name

    @text_font.setter
    def text_font(self, font: str = None) -> None:
        """
        changes the font style

        Parameters
        ----------
            font : str
                name of the font if system path otherwise
        """
        name, path = None, None
        if "/" in font or "\\" in font or "." in font or ".ttf" in font:
            path = font
        else:
            name = font

        if name is None and path is None:
            warn(
                f"WARNING [renderer] : blank instruction at text_font setter, nothing changed"
            )
            return

        size = self._text_size
        if name is not None:
            self._ff_name = name
            self._ff_is_sys = True
            self.font = pygame.font.SysFont(self._ff_name, size)
        elif path is not None:
            self._ff_name = path
            self._ff_is_sys = False
            self.font = pygame.font.Font(self._ff_name, size)

    @property
    def text_color(self) -> tuple[int, int, int]:
        """
        gets the current text color
        """
        return self._text_color

    @text_color.setter
    def text_color(self, color: Union[tuple[int, int, int], int, str]) -> None:
        """
        changes the text color globally

        Parameters
        ----------
            color : tuple | int | str
                the new color
        """
        if isinstance(color, tuple) and len(color) == 3:
            self._text_color = color
        elif isinstance(color, int):
            self._text_color = color, color, color
        elif isinstance(color, str):
            try:
                self._text_color = COLORS[color.lower()]
            except KeyError:
                close = difflib.get_close_matches(color.lower(),
                                                  COLORS.keys(),
                                                  n=1,
                                                  cutoff=.5)[0]
                warn(
                    f"ERROR [renderer] : {color} is not a valid color name, using closest match {close} instead"
                )
                self._text_color = COLORS[close]
        else:
            warn(
                f"ERROR [renderer] : {color} not a valid color parameter, nothing changed"
            )

    @property
    def pixels(self) -> pygame.PixelArray:
        """
        gets the current pixel array
        """
        return self._pixels

    def begin_shape(self) -> None:
        """
        begins drawing shape\\
        use ``end_shape`` with additional args to end drawing shape
        """
        if self._is_drawing_shape:
            warn(
                f"ERROR [renderer] : already drawing a shape, nothing happened"
            )
            return
        self._is_drawing_shape = True

    def end_shape(self, filled: bool = False, closed: bool = False) -> None:
        """
        ends drawing shape\\
        use ``begin_shape`` to begin drawing shape

        Parameters
        ----------
            filled : bool, (optional)
                if shape is filled (will not enable filling)
                defaults to False
            closed : bool, (optional)
                if first point connected to last
                defaults to False
        """
        if not self._is_drawing_shape:
            warn(f"ERROR [renderer] : not drawing shape, nothing happened")
            return
        if closed:
            self._all_vertexes.append(self._all_vertexes[0])

        if filled:
            self.polygon(*self._all_vertexes)
        elif not filled:
            self.lines(*self._all_vertexes, closed=False)

        self._is_drawing_shape = False
        self._all_vertexes = []

    def vertex(self, point: Union[tuple, list, Vector]):
        """
        draws shapes with given vertexes

        Parameters
        ----------
            point : tuple | list | Vector
                a point
        """
        if not self._is_drawing_shape:
            warn(f"ERROR [renderer] : not drawing shape, nothing happened")
            return
        self._all_vertexes.append(point)

    def _debug_enabled_drawing_methods(self) -> None:
        """
        Enables stroking if both stroking and filling are disabled\\
        Resets stroke weight back to 1 if needed as 0 fills all shapes
        """
        if (not self._stroke) and (not self._fill):
            warn(
                f"ERROR [renderer] : stroking and filling both set to False, stroking is now enabled"
            )
            self._stroke = True
            if self.stroke_weight <= 0:
                warn(
                    f"ERROR [renderer] : stroke weight set to {self.stroke_weight}, stroke weight is now 1"
                )
                self.stroke_weight = 1

    def _offset_point(self, point: tuple) -> list[float, float]:
        """
        Offsets a point based on _x_offset and _y_offset

        Parameters
        ----------
            point : tuple
                the point to apply the transformation

        Returns
        -------
            list : transformed point
        """
        point = list(point)
        new_point = [point[0] + self._x_offset, point[1] + self._y_offset]
        return new_point

    def _rotate_point(self, point: tuple) -> list[float, float]:
        """
        Rotates a point based on _rot_angle in radians\\
        relatives to the axis origin

        Parameters
        ----------
            point : tuple
                the point to apply the transformation

        Returns
        -------
            list : transformed point
        """
        point = list(point)
        new_point = [
            point[0] * m.cos(self._rot_angle) -
            point[1] * m.sin(self._rot_angle),
            point[0] * m.sin(self._rot_angle) +
            point[1] * m.cos(self._rot_angle)
        ]
        return new_point

    def _scale_point(self, point: tuple) -> list[float, float]:
        """
        Scales a point based on _scale_factor\\
        relatives to the axis origin

        Parameters
        ----------
            point : tuple
                the point to apply the transformation

        Returns
        -------
            list : transformed point
        """
        point = list(point)
        new_point = [
            point[0] * self._scale_factor, point[1] * self._scale_factor
        ]
        return new_point

    def line(self, point1: Union[tuple, list, Vector],
             point2: Union[tuple, list, Vector]) -> None:
        """
        draws a line on the screen\\
        uses the stroke color even if stroking is disabled

        Parameters
        ----------
            point1 : tuple | list | Vector
                first point
            point2 : tuple | list | Vector
                second point
        """
        if self._has_scale:
            point1 = self._scale_point(point1)
            point2 = self._scale_point(point2)
        if self._has_rotation:
            point1 = self._rotate_point(point1)
            point2 = self._rotate_point(point2)
        if self._has_translation:
            point1 = self._offset_point(point1)
            point2 = self._offset_point(point2)

        color = self.stroke
        weight = self.stroke_weight
        pygame.draw.line(self._window, color, point1[:2], point2[:2], weight)

    def aaline(self, point1: Union[tuple, list, Vector],
               point2: Union[tuple, list, Vector]) -> None:
        """
        draws an anti-aliased line of thickness 1 on the screen\\
        uses the stroke color even if stroking is disabled

        Parameters
        ----------
            point1 : tuple | list | Vector
                first point
            point2 : tuple | list | Vector
                second point
        """
        if self._has_scale:
            point1 = self._scale_point(point1)
            point2 = self._scale_point(point2)
        if self._has_rotation:
            point1 = self._rotate_point(point1)
            point2 = self._rotate_point(point2)
        if self._has_translation:
            point1 = self._offset_point(point1)
            point2 = self._offset_point(point2)

        color = self.stroke
        weight = self.stroke_weight
        pygame.draw.aaline(self._window, color, point1[:2], point2[:2], weight)

    def lines(self,
              *points: Union[tuple, list, Vector],
              closed: bool = True) -> None:
        """
        draws lines on the screen\\
        uses the stroke color even if stroking is disabled

        Parameters
        ----------
            points : tuples | lists | Vectors
                each additional arg is a point
            closed : bool, (optional)
                last point connected to first
                defaults to True
        """
        points: list[list[float, float]] = [list(p[:2]) for p in points]
        if self._has_scale:
            points = list(map(self._scale_point, points))
        if self._has_rotation:
            points = list(map(self._rotate_point, points))
        if self._has_translation:
            points = list(map(self._offset_point, points))

        color = self.stroke
        pygame.draw.lines(self._window, color, closed, points)

    def aalines(self,
                *points: Union[tuple, list, Vector],
                closed: bool = True) -> None:
        """
        draws anti-aliased lines of thickness 1 on the screen\\
        uses the stroke color even if stroking is disabled

        Parameters
        ----------
            points : tuples | lists | Vectors
                each additional arg is a point
            closed : bool, (optional)
                last point connected to first
                defaults to True
        """
        points: list[list[float, float]] = [list(p[:2]) for p in points]
        if self._has_scale:
            points = list(map(self._scale_point, points))
        if self._has_rotation:
            points = list(map(self._rotate_point, points))
        if self._has_translation:
            points = list(map(self._offset_point, points))

        color = self.stroke
        pygame.draw.aalines(self._window, color, closed, points)

    def polygon(self, *points: Union[tuple, list, Vector]) -> None:
        """
        draws a polygon on the screen\\
        calls debug_enabled_drawing_methods first

        Parameters
        ----------
            points : tuples | lists | Vectors
                each additional arg is a point
        """
        self._debug_enabled_drawing_methods()
        points: list[list[float, float]] = [list(p[:2]) for p in points]
        if self._has_scale:
            points = list(map(self._scale_point, points))
        if self._has_rotation:
            points = list(map(self._rotate_point, points))
        if self._has_translation:
            points = list(map(self._offset_point, points))

        # fill
        if self._fill:
            pygame.draw.polygon(self._window, self.fill, points, 0)
        # stroke
        if self._stroke:
            pygame.draw.polygon(self._window, self.stroke, points,
                                self.stroke_weight)

    def rect(self, point: Union[tuple, list, Vector], width: int,
             height: int) -> None:
        """
        draws a rectangle on the screen\\
        calls debug_enabled_drawing_methods first

        Parameters
        ----------
            point : tuple | list | Vector
                base point of the rectangle
            width : int
                the width of the rectangle
            height : int
                the height of the rectangle
        """
        self._debug_enabled_drawing_methods()
        point = list(point)
        if self.rect_mode == CENTER:
            point[0] -= width // 2
            point[1] -= height // 2
        # if self._has_rotation:
        #     point = self._rotate_point(point)
        # if self._has_translation:
        #     point = self._offset_point(point)

        x, y = point[:2]
        points = [(x, y), (x + width, y), (x + width, y + height),
                  (x, y + height)]

        if self._has_scale:
            points = list(map(self._scale_point, points))
        if self._has_rotation:
            points = list(map(self._rotate_point, points))
        if self._has_translation:
            points = list(map(self._offset_point, points))

        # # fill
        # if self._fill:
        #     pygame.draw.rect(self._window, self.fill, (point[:2], (width, height)), 0)
        # # stroke
        # if self._stroke:
        #     pygame.draw.rect(self._window, self.stroke, (point[:2], (width, height)), self.stroke_weight)

        # fill
        if self._fill:
            pygame.draw.polygon(self._window, self.fill, points, 0)
        # stroke
        if self._stroke:
            pygame.draw.polygon(self._window, self.stroke, points,
                                self.stroke_weight)

    def square(self, point: Union[tuple, list, Vector], size: int) -> None:
        """
        draws a square on the screen\\
        calls debug_enabled_drawing_methods first

        Parameters
        ----------
            point : tuple | list | Vector
                base point of the rectangle
            size : int
                the size of the square
        """
        self._debug_enabled_drawing_methods()
        point = list(point)
        if self.rect_mode == CENTER:
            point[0] -= size // 2
            point[1] -= size // 2
        # if self._has_rotation:
        #     point = self._rotate_point(point)
        # if self._has_translation:
        #     point = self._offset_point(point)

        x, y = point[:2]
        points = [(x, y), (x + size, y), (x + size, y + size), (x, y + size)]

        if self._has_scale:
            points = list(map(self._scale_point, points))
        if self._has_rotation:
            points = list(map(self._rotate_point, points))
        if self._has_translation:
            points = list(map(self._offset_point, points))

        # # fill
        # if self._fill:
        #     pygame.draw.rect(self._window, self.fill, (point[:2], (size, size)), 0)
        # # stroke
        # if self._stroke:
        #     pygame.draw.rect(self._window, self.stroke, (point[:2], (size, size)), self.stroke_weight)

        # fill
        if self._fill:
            pygame.draw.polygon(self._window, self.fill, points, 0)
        # stroke
        if self._stroke:
            pygame.draw.polygon(self._window, self.stroke, points,
                                self.stroke_weight)

    def ellipse(self, point: Union[tuple, list, Vector], width: int,
                height: int) -> None:
        """
        draws an ellipse on the screen\\
        calls debug_enabled_drawing_methods first

        Parameters
        ----------
            point : tuple | list | Vector
                base point of the rectangle for the ellipse
            width : int
                the width of the rectangle for the ellipse
            height : int
                the height of the rectangle for the ellipse
        """
        self._debug_enabled_drawing_methods()
        point = list(point)
        if self.rect_mode == CENTER:
            point[0] -= width // 2
            point[1] -= height // 2
        if self._has_scale:
            point = self._scale_point(point)
            width *= self._scale_factor
            height *= self._scale_factor
        if self._has_rotation:
            point = self._rotate_point(point)
        if self._has_translation:
            point = self._offset_point(point)

        # fill
        if self._fill:
            pygame.draw.ellipse(self._window, self.fill,
                                (point[:2], (width, height)), 0)
        # stroke
        if self._stroke:
            pygame.draw.ellipse(self._window, self.stroke,
                                (point[:2], (width, height)),
                                self.stroke_weight)

    def circle(self, center: Union[tuple, list, Vector], radius: int) -> None:
        """
        draws a circle on the screen\\
        calls debug_enabled_drawing_methods first

        Parameters
        ----------
            center : tuple | list | Vector
                center point of the circle
            radius : int
                circle radius
        """
        self._debug_enabled_drawing_methods()
        if self._has_scale:
            center = self._scale_point(center)
            radius *= self._scale_factor
        if self._has_rotation:
            center = self._rotate_point(center)
        if self._has_translation:
            center = self._offset_point(center)

        # fill
        if self._fill:
            pygame.draw.circle(self._window, self.fill, center[:2], radius, 0)
        # stroke
        if self._stroke:
            pygame.draw.circle(self._window, self.stroke, center[:2], radius,
                               self._stroke_weight)

    def arc(self, point: Union[tuple, list, Vector], width: int, height: int,
            start: float, stop: float) -> None:
        """
        draws an arc on the screen\\
        calls debug_enabled_drawing_methods first

        Parameters
        ----------
            point : tuple | list | Vector
                base point of the rectangle for the arc
            width : int
                the width of the rectangle for the arc
            height : int
                the height of the rectangle for the arc
            start : float
                start angle for the arc
            stop : float
                stop angle for the arc
        """
        self._debug_enabled_drawing_methods()
        point = list(point)
        if self.rect_mode == CENTER:
            point[0] -= width // 2
            point[1] -= height // 2
        if self._has_scale:
            point = self._scale_point(point)
            width *= self._scale_factor
            height *= self._scale_factor
        if self._has_rotation:
            point = self._rotate_point(point)
            start += self._rot_angle
            stop += self._rot_angle
        if self._has_translation:
            point = self._offset_point(point)

        # fill
        if self._fill:
            pygame.draw.ellipse(self._window, self.fill,
                                (point[:2], (width, height)), start, stop, 0)
        # stroke
        if self._stroke:
            pygame.draw.ellipse(self._window, self.stroke,
                                (point[:2], (width, height)), start, stop,
                                self.stroke_weight)

    def point(self, point: Union[tuple, list, Vector]) -> None:
        """
        draws a point on the screen\\
        uses stroke color and stroke weight even if stroking is disabled

        Parameters
        ----------
            point : tuple | list | Vector
                the point coordinates
        """
        if self._has_scale:
            point = self._scale_point(point)
        if self._has_rotation:
            point = self._rotate_point(point)
        if self._has_translation:
            point = self._offset_point(point)

        pygame.draw.circle(self._window, self.stroke, point[:2],
                           self._stroke_weight, 0)

    def sprites(self, group: pygame.sprite.Group) -> None:
        """
        draws a group of sprites on the screen\\
        uses rect attribute as coordinates

        Parameters
        ----------
            group : pygame.sprite.Group
                a group of sprites
        """
        group.draw(self._window)

    def load_image(self, path: str) -> pygame.Surface:
        """
        loads an image
        """
        image = pygame.image.load(path)
        return image

    def transform_image(self,
                        image: pygame.Surface,
                        scale: float = 1,
                        angle: float = 0) -> pygame.Surface:
        """
        applies scale and / or rotation on a image and returns a new image\\
        does not modify the original image
        """
        new_image = pygame.transform.rotozoom(image, angle, scale)
        return new_image

    def get_image_size(self, image: pygame.Surface) -> tuple[int, int]:
        """
        gets width, height of an image
        """
        return image.get_width(), image.get_height()

    def draw_image(self, x: int, y: int, image: pygame.Surface) -> None:
        """
        displays an image on the screen

        Parameters
        ----------
            x : int
                x-coordinate
            y : int
                y-coordinate
            image : pygame.Surface
                loaded image
        """
        if self.rect_mode == CENTER:
            dx, dy = self.get_image_size()
            x -= dx / 2
            y -= dy / 2
        self._window.blit(image, (x, y))

    def text(self, x: int, y: int, text: str) -> None:
        """
        displays some text on the screen\\
        uses text color and size

        Parameters
        ----------
            x : int
                x-coordinate
            y : int
                y-coordinate
            text : str
                the text to display
        """
        color = self.text_color
        text_label = self.font.render(text, True, color)
        self._window.blit(text_label, (x, y))

    def background(self, *color: Union[int, str]) -> None:
        """
        fills the screen with a unique color

        Parameters
        ----------
            color : tuple[int, int, int] | int | str
                color to fill the screen with
        """
        if len(color) == 1:
            color = color[0]
        if isinstance(color, int):
            color = color, color, color
        elif isinstance(color, tuple) and len(color) == 3:
            pass
        elif isinstance(color, str):
            try:
                color = COLORS[color.lower()]
            except KeyError:
                close = difflib.get_close_matches(color.lower(),
                                                  COLORS.keys(),
                                                  n=1,
                                                  cutoff=.5)[0]
                warn(
                    f"ERROR [renderer] : {color} is not a valid color name, using closest match {close} instead"
                )
                color = COLORS[close]
        else:
            warn(
                f"ERROR [renderer] : {color} not a valid color parameter, applaying default dark background"
            )
            color = 51, 51, 51
        self._bg = color
        self._window.fill(color)

    def translate(self, x: float = 0, y: float = 0) -> None:
        """
        translates the axes origin, additive

        Parameters
        ----------
            x : float
                translation for x-axis
            y : float
                translation for y-axis
        """
        if self._has_scale:
            x, y = self._scale_point((x, y))
        if self._has_rotation:
            x, y = self._rotate_point((x, y))

        self._x_offset += x
        self._y_offset += y
        if self._x_offset == 0 and self._y_offset == 0:
            self._has_translation = False
        else:
            self._has_translation = True

    def rotate(self, angle: float) -> None:
        """
        rotates the axes around the axis origin, additive

        Parameters
        ----------
            angle : float
                angle in randians
        """
        self._rot_angle += angle
        if self._rot_angle == 0:
            self._has_rotation = False
        else:
            self._has_rotation = True

    def rotate_display(self, angle: float) -> None:
        """
        rotates the entire window by some angle, relative to the center of the screen\\
        affects what has already been drawn only

        Parameters
        ----------
            angle : float
                angle in radians
        """
        surface = pygame.transform.rotate(self._window, m.degrees(angle))
        x = (self.win_width - surface.get_width()) / 2
        y = (self.win_height - surface.get_height()) / 2
        self._window.blit(surface, (x, y))

    def scale(self, scale: float) -> None:
        """
        scales what will be drawn on the window\\
        does not affect the stroke weight, additive

        Parameters
        ----------
            scale : float
                scale factor, must be greater than 0
        """
        if scale <= 0:
            warn(
                f"WARNING [renderer] : scale of {scale} is not allowed, nothing happened"
            )
            return
        self._scale_factor *= scale
        if self._scale_factor == 1:
            self._has_scale = False
        else:
            self._has_scale = True

    def scale_display(self, scale: float) -> None:
        """
        scales the entire window by some amount, relative to the center of the screen\\
        affects what has already been drawn only

        Parameters
        ----------
            scale : float
                scale (zoom factor, must be positive)
        """
        surface = pygame.transform.rotozoom(self._window, 0, scale)
        x = (self.win_width - surface.get_width()) / 2
        y = (self.win_height - surface.get_height()) / 2
        self._window.blit(surface, (x, y))

    def reset_matrix(self) -> None:
        """
        reset all translations and drawing modes back to original\\
        almost as if the renderer could pop to its original state\\
        does not affect colors and sizes
        """
        self._reset_translation()
        self._reset_rotation()
        self._reset_scale()
        self.rect_mode = CORNER
        self.translation_behavior = RESET
        self.rotation_behavior = RESET
        self.scale_behavior = RESET

    def _reset_translation(self) -> None:
        """
        resets the axis origin back to normal
        """
        self._has_translation = False
        self._x_offset = 0
        self._y_offset = 0

    def _reset_rotation(self) -> None:
        """
        resets rotation angle back to 0
        """
        self._has_rotation = False
        self._rot_angle = 0

    def _reset_scale(self) -> None:
        """
        resets scale factor back to 1
        """
        self._has_scale = False
        self._scale_factor = 1

    @property
    def translation_behavior(self) -> str:
        """
        gets the current translation behavior
        """
        return self._translation_behavior

    @translation_behavior.setter
    def translation_behavior(self, behavior: str) -> None:
        """
        sets the global translation behavior

        Parameters
        ----------
            behavior : str
                KEEP | RESET
        """
        if behavior not in ("RESET", "KEEP"):
            warn(
                f"WARNING [renderer] : {behavior} is not a valid translation behavior, nothing happened"
            )
            return
        self._translation_behavior = behavior

    @property
    def rotation_behavior(self) -> str:
        """
        gets the current rotation behavior
        """
        return self._rotation_behavior

    @rotation_behavior.setter
    def rotation_behavior(self, behavior: str) -> None:
        """
        sets the global rotation behavior

        Parameters
        ----------
            behavior : str
                KEEP | RESET
        """
        if behavior not in ("RESET", "KEEP"):
            warn(
                f"WARNING [renderer] : {behavior} is not a valid rotation behavior, nothing happened"
            )
            return
        self._rotation_behavior = behavior

    @property
    def scale_behavior(self) -> str:
        """
        gets the global scale behavior
        """
        return self._scale_behavior

    @scale_behavior.setter
    def scale_behavior(self, behavior: str) -> None:
        """
        sets the global scale behavior

        Parameters
        ----------
            behavior : str
                KEEP | RESET
        """
        if behavior not in ("RESET", "KEEP"):
            warn(
                f"WARNING [renderer] : {behavior} is not a valid scale behavior, nothing happened"
            )
            return
        self._scale_behavior = behavior

    def _add_button(self, button: Button) -> None:
        """
        adds a new button to the sprite Group

        Parameters
        ----------
            button : Button
                new button
        """
        self._all_buttons.add(button)

    def create_button(self, x: int, y: int, name: str, **kwargs) -> Button:
        """
        creates a new button and adds it to the sprite Group

        Parameters
        ----------
            x : int
                the x-coordinate of the button
            y : int
                the y-coordinate of the button
            name : str
                the name of the button

        Options
        -------
            count : int
                number of frames to pass while un-clicked to be able to trigger the button again
            action : python function
                the action to trigger when pressed
            height : int
                the height of the sprite
            width : int
                the width of the sprite
            shape : str
                the shape of the button
                RECTANGLE | ELLIPSE
            color : tuple | int | str
                color to fill button and enables filling
                if both color and stroke are None, the button will be filled by default
            stroke : str
                color to draw the outside box and enables stroking
            weight : int
                stroke weight if stroke is not None

        Returns
        -------
            Button : gets the new button if successfull
        """
        button = Button(self, x, y, name, **kwargs)
        if button.has_error:
            return
        self._add_button(button)
        self._has_buttons = True
        return button

    def _remove_button(self, button: Button) -> None:
        """
        kills the living sprite (deals with type None)\\
        removes it from sprite Group but still can be accessed

        Parameters
        ----------
            button : Button
                the button to remove
        """
        if button is not None:
            self._all_buttons.remove(button)
            if len(self._all_buttons) == 0:
                self._has_buttons = False

    def get_button(self, name: str) -> Button:
        """
        gets a button based on its name\\
        does nothing if not matched

        Parameters
        ----------
            name : str
                name of the button to get

        Returns
        -------
            Button : matching button
        """
        sprite = None
        found = 0
        for button in self._all_buttons:
            if button.name == name:
                sprite = button
                found += 1
        if found == 0:
            warn("WARNING [renderer] : no matching button")
        elif found >= 2:
            warn(
                "WARNING [renderer] : to many matches - considering last found"
            )
        return sprite

    def kill_button(self, name: str) -> None:
        """
        kills a button based on its name\\
        does nothing if not matched\\
        does not return anything

        Parameters
        ----------
            name : str
                name of the button to remove
        """
        sprite = None
        found = 0
        for button in self._all_buttons:
            if button.name == name:
                sprite = button
                found += 1
        if found == 0:
            warn("WARNING [renderer] : no matching button")
        elif found >= 2:
            warn(
                "WARNING [renderer] : to many matches - considering last found"
            )
        self._remove_button(sprite)

    def pop_button(self, name: str) -> Button:
        """
        kills a button based on its name\\
        returns None if not matched\\
        else returns the matching button

        Parameters
        ----------
            name : str
                name of the button to remove

        Returns
        -------
            Button | None : matched button if found
        """
        sprite = None
        found = 0
        for button in self._all_buttons:
            if button.name == name:
                sprite = button
                found += 1
        if found == 0:
            warn("WARNING [renderer] : no matching button")
        elif found >= 2:
            warn(
                "WARNING [renderer] : to many matches - considering last found"
            )
        self._remove_button(sprite)
        return sprite

    def _add_slider(self, slider: Slider) -> None:
        """
        adds a new slider to the sprite Group

        Parameters
        ----------
            slider : Slider
                new slider
        """
        self._all_sliders.add(slider)

    def create_slider(self, x: int, y: int, name: str, min: float, max: float,
                      value: float, incr: int, **kwargs) -> Slider:
        """
        creates a new slider and adds it to the sprite group

        Parameters
        ----------
            x : int
                x-coordinate
            y : int
                y-coordinate
            name : str
                name of the slider
            min : float
                minimum value
            max : float
                maximum value
            value : float
                current value
            incr : int
                rounding (may not be effective depending on the length of the slider)

        Options
        -------
            radius : int
                the radius of the slider cursor
            shape : str
                shape of the slider cursor
                SQUARE | CIRCLE | CROSS | PLUS
            thickness : int
                the thickness of the slider bar
            color : tuple | int | str
                default color of the bar
            fullcolor : tuple | int | str
                color of the bar when its full
            length : int
                the length of the slider bar
            count : int
                number of frames to pass while inactive to send value

        Returns
        -------
            Slider : gets the new slider if successfull
        """
        slider = Slider(self, x, y, name, min, max, value, incr, **kwargs)
        if slider.has_error:
            return
        self._add_slider(slider)
        self._has_sliders = True
        return slider

    def _remove_slider(self, slider: Slider) -> None:
        """
        kills the living sprite (deals with type None)\\
        removes it from sprite Group but still can be accessed

        Parameters
        ----------
            slider : Slider
                the slider to remove
        """
        if slider is not None:
            self._all_sliders.remove(slider)
            if len(self._all_sliders) == 0:
                self._has_sliders = False

    def get_slider(self, name: str) -> Slider:
        """
        gets a slider based on its name\\
        does nothing if not matched

        Parameters
        ----------
            name : str
                name of the slider to get

        Returns
        -------
            Slider : matching slider
        """
        sprite = None
        found = 0
        for slider in self._all_sliders:
            if slider.name == name:
                sprite = slider
                found += 1
        if found == 0:
            warn("WARNING [renderer] : no matching slider")
        elif found >= 2:
            warn(
                "WARNING [renderer] : to many matches - considering last found"
            )
        return sprite

    def kill_slider(self, name: str) -> None:
        """
        kills a slider based on its name\\
        does nothing if not matched\\
        does not return anything

        Parameters
        ----------
            name : str
                name of the slider to remove
        """
        sprite = None
        found = 0
        for slider in self._all_sliders:
            if slider.name == name:
                sprite = slider
                found += 1
        if found == 0:
            warn("WARNING [renderer] : no matching slider")
        elif found >= 2:
            warn(
                "WARNING [renderer] : to many matches - considering last found"
            )
        self._remove_slider(sprite)

    def pop_slider(self, name: str) -> Button:
        """
        kills a slider based on its name\\
        returns None if not matched\\
        else returns the matching slider

        Parameters
        ----------
            name : str
                name of the slider to remove

        Returns
        -------
            Slider | None : matched slider if found
        """
        sprite = None
        found = 0
        for slider in self._all_sliders:
            if slider.name == name:
                sprite = slider
                found += 1
        if found == 0:
            warn("WARNING [renderer] : no matching slider")
        elif found >= 2:
            warn(
                "WARNING [renderer] : to many matches - considering last found"
            )
        self._remove_slider(sprite)
        return sprite

    def get_slider_value(self, name: str) -> float:
        """
        gets the value of a slider based on its name\\
        may not return anything if the slider is unmatched

        Parameters
        ----------
            name : str
                the name of the slider

        Returns
        -------
            float | None : the current value of the slider
        """
        sprite = None
        found = 0
        for slider in self._all_sliders:
            if slider.name == name:
                sprite = slider
                found += 1
        if found == 0:
            warn("WARNING [renderer] : no matching slider")
            return
        if found >= 2:
            warn(
                "WARNING [renderer] : to many matches - considering last found"
            )
        return sprite.value

    def _add_menu(self, menu: Menu) -> None:
        """
        adds a new menu to the sprite Group

        Parameters
        ----------
            menu : Menu
                new menu
        """
        self._all_menus.add(menu)

    def create_menu(self, name: str, **kwargs) -> Menu:
        """
        creates a new menu and adds it to the sprite group

        Parameters
        ----------
            name : str
                name of the menu

        Options
        -------
            side : str
                side of the window to display the menu
                LEFT | RIGHT
            length : int
                length of the menu (its height)
                by default the menu height will be on auto
            background : None | bool | tuple | int | str
                draw a background when expanded
            color : tuple | int | str
                lines color used for drawing when menu is visible
            text_color : tuple | int | str
                text color used for display text inside the menu
            text_size : int
                text size of the menu

        Keywords Arguments
        ------------------
            * : str\\
                name of the buttons on the menu in order\\
                must be linked to a python function

        Returns
        -------
            Menu : gets the new menu if successfull
        """
        menu = Menu(self, name, **kwargs)
        if menu.has_error:
            return
        if (menu.side == LEFT
                and self._has_left_menu) or (menu.side == RIGHT
                                             and self._has_right_menu):
            warn(
                f"ERROR [renderer] : already have a {menu.side} menu, try again by changing side"
            )
            return
        self._add_menu(menu)
        if menu.side == LEFT:
            self._has_left_menu = True
        elif menu.side == RIGHT:
            self._has_right_menu = True
        return menu

    def _remove_menu(self, menu: Menu) -> None:
        """
        kills the living sprite (deals with type None)\\
        removes it from sprite Group but still can be accessed

        Parameters
        ----------
            menu : Menu
                the menu to remove
        """
        if menu is not None:
            self._all_sliders.remove(menu)
            if len(self._all_sliders) == 0:
                if menu.side == LEFT:
                    self._has_left_menu = False
                elif menu.side == RIGHT:
                    self._has_right_menu = False

    def get_menu(self, name: str) -> Menu:
        """
        gets a menu based on its name\\
        does nothing if not matched

        Parameters
        ----------
            name : str
                name of the menu to get

        Returns
        -------
            Menu : matching menu
        """
        sprite = None
        found = 0
        for menu in self._all_menus:
            if menu.name == name:
                sprite = menu
                found += 1
        if found == 0:
            warn(f"WARNING [renderer] : no matching menu")
        elif found >= 2:
            warn(
                f"WARNING [renderer] : to many matches - considering last found"
            )
        return sprite

    def kill_menu(self, name: str) -> None:
        """
        kills a menu based on its name\\
        does nothing if not matched\\
        does not return anything

        Parameters
        ----------
            name : str
                name of the menu to remove
        """
        sprite = None
        found = 0
        for menu in self._all_menus:
            if menu.name == name:
                sprite = menu
                found += 1
        if found == 0:
            warn(f"WARNING [renderer] : no matching menu")
        elif found >= 2:
            warn(
                f"WARNING [renderer] : to many matches - considering last found"
            )
        self._remove_menu(sprite)

    def pop_menu(self, name: str) -> Menu:
        """
        kills a menu based on its name\\
        returns None if not matched\\
        else returns the matching menu
        
        Parameters
        ----------
            name : str
                name of the menu to remove

        Returns
        -------
            Menu | None : matching menu if found
        """
        sprite = None
        found = 0
        for menu in self._all_menus:
            if menu.name == name:
                sprite = menu
                found += 1
        if found == 0:
            warn(f"WARNING [renderer] : no matching menu")
        elif found >= 2:
            warn(
                f"WARNING [renderer] : to many matches - considering last found"
            )
        self._remove_menu(sprite)
        return sprite

    def _add_scrollbar(self, scrollbar: ScrollBar) -> None:
        """
        sets a new scrollbar

        Parameters
        ----------
            scrollbar : ScrollBar
                new scrollbar
        """
        self._scrollbar = scrollbar

    def create_scrollbar(self, mn: int, mx: int, **kwargs) -> ScrollBar:
        """
        creates a scrollbar and adds it to the renderer

        Parameters
        ----------
            mn : int
                minimum y view point
            mx : int
                maximum y view point

        Options
        -------
            color1 : Union[tuple[int, int, int], int, str], (optional)
                color of the scrollbar item
            color2 : Union[tuple[int, int, int], int, str], (optional)
                color of the background when activated
        """
        scrollbar = ScrollBar(self, (mn, mx), **kwargs)
        if scrollbar.has_error:
            return
        if self._has_scrollbar:
            warn(
                f"ERROR [renderer] : already have a scrollbar, nothing happened"
            )
            return
        self._add_scrollbar(scrollbar)
        self._has_scrollbar = True
        return scrollbar

    def _remove_scrollbar(self) -> None:
        """
        kills the living sprite\\
        removes it but still can be accessed
        """
        if self._scrollbar is not None:
            self._scrollbar = None
            self._has_scrollbar = False

    def get_scrollbar(self) -> ScrollBar:
        """
        gets active scrollbar

        Returns
        -------
            ScrollBar : unique scrollbar if found
        """
        return self._scrollbar

    def kill_scrollbar(self) -> None:
        """
        kills living scrollbar\\
        does not return anything
        """
        self._remove_scrollbar()

    def pop_scrollbar(self) -> ScrollBar:
        """
        kills living scrollbar

        Returns
        -------
            ScrollBar | None : unique scrollbar
        """
        sprite = self._scrollbar
        self._remove_scrollbar()
        return sprite

    @staticmethod
    def _events() -> list:
        """
        gets all events
        """
        return pygame.event.get()

    def _quit_check(self) -> None:
        """
        loop trough events and look for the QUIT event\\
        does nothing if found\\
        must not be called if Renderer._events() is called before
        """
        for event in self._events():
            if event.type == pygame.QUIT:
                self._is_running = False

    @property
    def fps(self) -> float:
        """
        gets real fps
        """
        return self._clock.get_fps()

    @fps.setter
    def fps(self, frames: int) -> None:
        """
        sets the frame rate of the screen\\
        setting this to a negative value will unlock the frame rate

        Parameters
        ----------
            frames : int
                frames per second
        """
        self._fps = frames

    def push(self) -> None:
        """
        adds the state of the renderer to the stack\\
        does not affect outer objects\\
        use ``pop`` to reset the state
        """
        self._save.append([
            self.fill, self._fill, self.stroke, self.stroke_weight,
            self._stroke, self._has_translation, self._x_offset,
            self._y_offset, self._has_rotation, self._rot_angle,
            self._has_scale, self._scale_factor, self.rect_mode,
            self.translation_behavior, self.rotation_behavior,
            self.scale_behavior, self.text_color, self.text_size
        ])
        self._has_save = True

    def pop(self) -> None:
        """
        resets the state of the renderer based on the previous save\\
        does nothing if save is not found\\
        use ``push`` to generate save
        """
        if not self._has_save:
            warn("WARNING [renderer] : no save was found, nothing changed")
            return
        self.fill, self._fill, self.stroke, self.stroke_weight, self._stroke, self._has_translation,self._x_offset, self._y_offset, self._has_rotation, self._rot_angle, self._has_scale, self._scale_factor,self.rect_mode, self.translation_behavior, self.rotation_behavior, self.scale_behavior, self.text_color, self.text_size =\
            self._save.pop()
        self._has_save = len(self._save) >= 1

    @property
    def key_binding(self) -> dict[int, int]:
        """
        gets current state of the Renderer key binding\\
        dict keys are keyboard keys identifiers (also used by pygame)\\
        dict values are indexes for bound function (actions property)
        """
        return self._key_binding

    def new_keypress(self,
                     key: int,
                     action: Callable[[], None],
                     behavior: str = PRESSED) -> None:
        """
        adds a new key and its corresponding action\\
        please use ``Renderer.keys.`` to find keys

        Parameters
        ----------
            key : int
                keyboard key identifier
            action : python function
                action to perform each time the key is pressed
            behavior : str, (optional)
                key behavior
                PRESSED | RELEASED | HOLD
                defaults to PRESSED
        """
        if key in self.key_binding:
            warn(
                f"ERROR [renderer] : {key} is already assigned to a function, try update_key instead"
            )
            return
        if behavior not in (PRESSED, RELEASED, HOLD):
            warn(
                f"ERROR [renderer] : {behavior} is not a valid key behavior, nothing happened"
            )
            return
        self._actions.append(action)
        self._keys_behavior.append(behavior)
        self._key_binding[key] = len(self._actions) - 1

    def update_keypress(self,
                        key: int,
                        action: Callable[[], None],
                        behavior: str = None) -> None:
        """
        updates the action of a given key\\
        please use ``Renderer.keys.`` to find keys

        Parameters
        ----------
            key : int
                keyboard key identifier
            action : python function
                new action
            behavior : str, (optional)
                key behavior (if not specified, will keep previous)
                PRESSED | RELEASED | HOLD
                defaults to None
        """
        if key not in self.key_binding:
            warn(
                f"ERROR [renderer] : {key} is not assigned to an existing function, try new_key instead"
            )
            return
        if behavior is not None and behavior not in (PRESSED, RELEASED, HOLD):
            warn(
                f"ERROR [renderer] : {behavior} is not a valid key behavior, nothing changed"
            )
            return
        i = self._key_binding[key]
        self._actions[i] = action
        if behavior is not None:
            self._keys_behavior[i] = behavior

    def kill_keypress(self, key: int) -> None:
        """
        removes the action of a given key\\
        changes the action to ``lambda: None`` and pops key binding\\
        leaves other keys action unchanged\\
        no action will be performed when the given key is pressed\\
        please use ``Renderer.keys.`` to find keys

        Parameters
        ----------
            key : int
                keyboard key identifier
        """
        if key not in self.key_binding:
            warn(
                f"ERROR [renderer] : {key} is not assigned to an existing function, can not kill"
            )
            return
        i = self._key_binding[key]
        self._actions[i] = lambda: None
        self._key_binding.pop(key)
        try:
            self._pressed.pop(key)
        except KeyError:
            pass

    def load_pixels(self) -> None:
        """
        loads pixels from the screen\\
        sets a pygame.PixelArray object accesible through property ``pixels``
        """
        if self._is_p_loaded:
            warn(
                "WARNING [renderer] : pixels are already loaded, nothing happened"
            )
            return
        self._is_p_loaded = True
        self._pixels = pygame.PixelArray(self._window)

    def update_pixels(self) -> None:
        """
        updates the pixel array\\
        use ``.load_pixels()`` to access the array
        """
        if not self._is_p_loaded:
            warn(
                "WARNING [renderer] : pixels are not loaded, nothing happened")
            return
        self._is_p_loaded = False
        self._pixels.close()

    def set_at(self, x: int, y: int) -> None:
        """
        sets a pixel at a given position\\
        uses stroke color even if stroking is disabled
        """
        color = self.stroke
        self._window.set_at((x, y), color)

    def flip(self) -> None:
        """
        updates window\\
        used for interractive drawing without the draw main loop
        """
        pygame.display.flip()

    def start(self) -> None:
        """
        opens a new window if the sketch is closed\\
        used for interractive drawing without the draw main loop
        """
        self._window = pygame.display.set_mode((self._width, self._height))
        pygame.display.set_caption(self._title)

    def quit(self) -> None:
        """
        quits the sketch by closing the window
        """
        pygame.quit()

    def set_bench_mode(self, bench: bool) -> None:
        """
        sets the bench mode of the Renderer\\
        setting this to True means nothing will be rendered except the basic drawings\\
        in other words, it skips all buttons, sliders, menus, ... management\\
        plus the Renderer no longer struggles to comply with frame rate (which are no longer relevant)

        Parameters
        ----------
            bench : bool
                True to not bother check for subclasses instances
        """
        if bench == self._benchmark:
            warn(
                f"WARNING [renderer] : bench mode is already at {bench}, nothing changed"
            )
            return
        self._benchmark = bench

    def set_background(self, *color: Union[None, int, str]) -> None:
        """
        sets an automated background every time through draw\\
        setting this to None will disable this feature

        Parameters
        ----------
            color : None | tuple[int, int, int] | int | str
                color to fill the screen with, might be ``None``
        """
        self._has_auto_bg = True

        if len(color) == 1:
            color = color[0]
        if color is None:
            self._has_auto_bg = False
            return

        if isinstance(color, int):
            color = color, color, color
        elif isinstance(color, tuple) and len(color) == 3:
            pass
        elif isinstance(color, str):
            try:
                color = COLORS[color.lower()]
            except KeyError:
                close = difflib.get_close_matches(color.lower(),
                                                  COLORS.keys(),
                                                  n=1,
                                                  cutoff=.5)[0]
                warn(
                    f"ERROR [renderer] : {color} is not a valid color name, using closest match {close} instead"
                )
                color = COLORS[close]
        else:
            warn(
                f"ERROR [renderer] : {color} not a valid color parameter, applaying default dark background"
            )
            color = 51, 51, 51
        self._bg = color

    def run(self,
            draw: Callable[[], None] = None,
            setup: Callable[[], None] = None) -> None:
        """
        the main loop of the program\\
        will call draw over and over until ``QUIT`` event is triggered

        The Renderer will try to load the ``draw`` and ``setup`` functions of the
        main file (the one that calls this method), and will raise an error if the
        draw function is not provided and can not be found in the file. However, you
        have the option to pass in the functions with the names of your likings as
        parameters to this method. In that case the Renderer will not scan your main
        file.

        Parameters
        ----------
            draw : python function, (optional)
                the main draw function
                defaults to None
            setup : python function, (optional)
                the setup function, will be called once
                defaults to None
        """
        if setup is None:
            if hasattr(__main__, "setup"):
                setup = __main__.setup
            else:
                setup = lambda: None
        if draw is None:
            if hasattr(__main__, "draw"):
                draw = __main__.draw
            else:
                raise NameError(
                    "ERROR [renderer] : draw function was not provided and was not found in the main file"
                )
        setup()

        while self._is_running:
            # drawing loop
            if self._has_auto_bg:
                self._window.fill(self._bg)
            draw()

            # translation, rotation, scale management
            if self.translation_behavior == RESET:
                self._reset_translation()
            if self.rotation_behavior == RESET:
                self._reset_rotation()
            if self.scale_behavior == RESET:
                self._reset_scale()

            # bench mode
            if self._benchmark:
                pygame.display.flip()
                for event in pygame.event.get():
                    if event.type == pygame.QUIT:
                        self._is_running = False
                        break
                continue

            # button management
            if self._has_buttons:
                for button in self._all_buttons:
                    if not button.is_hidden:
                        button.draw()

            # slider management
            if self._has_sliders:
                for slider in self._all_sliders:
                    if not slider.is_hidden:
                        slider.draw()

            # menu management
            if self._has_left_menu or self._has_right_menu:
                for menu in self._all_menus:
                    if not menu.is_hidden:
                        menu.draw()

            # scrollbar management
            if self._has_scrollbar:
                scrollbar = self._scrollbar
                if not scrollbar.is_hidden:
                    scrollbar.draw()
                    self._has_translation = True
                    value = scrollbar.value
                    if self.translation_behavior == RESET:
                        self._y_offset -= value
                    elif self.translation_behavior == KEEP:
                        self._y_offset += self._ps_value
                        self._y_offset -= value
                    self._ps_value = value

            # trigerring buttons, sliders and menus
            pos: tuple[int, int] = pygame.mouse.get_pos()
            if pygame.mouse.get_pressed()[0] != 0:

                if self._has_buttons:
                    for button in self._all_buttons:
                        if not button.is_hidden and button.collide(
                                pos) and button.check_click():
                            button.on_press()
                            button.reinit_click()

                if self._has_sliders:
                    for slider in self._all_sliders:
                        if not slider.is_hidden and slider.collide(pos):
                            slider.set_value(pos)
                            slider.reinit_click()

                if self._has_left_menu or self._has_right_menu:
                    for menu in self._all_menus:
                        if not menu.is_hidden and menu.check_click():
                            menu.update_state(pos)
                            if (i := menu.collide(pos)) is not None:
                                menu.trigger(i)
                            menu.reinit_click()

                if self._has_scrollbar:
                    scrollbar = self._scrollbar
                    if not scrollbar.is_hidden and scrollbar.collide(pos):
                        if not scrollbar.is_pinned():
                            scrollbar.set_pin(pos)
                        else:
                            scrollbar.set_value_by_y(pos[1] -
                                                     scrollbar.get_pin())

            else:
                if self._has_buttons:
                    for button in self._all_buttons:
                        button.click()
                if self._has_sliders:
                    for slider in self._all_sliders:
                        slider.click()
                if self._has_left_menu or self._has_right_menu:
                    for menu in self._all_menus:
                        menu.click()
                if self._has_scrollbar:
                    scrollbar = self._scrollbar
                    scrollbar.update_state(pos)
                    scrollbar.unpin()

            # quit event and keys
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    self._is_running = False

                if event.type == pygame.MOUSEBUTTONDOWN:
                    if self._has_scrollbar:
                        scrollbar = self._scrollbar
                        if event.button == 4:
                            scrollbar.scroll_up()
                        if event.button == 5:
                            scrollbar.scroll_down()

                if event.type == pygame.KEYUP:
                    if (k := event.key) in self.key_binding:
                        i = self._key_binding[k]

                        if self._keys_behavior[i] == RELEASED:
                            self._actions[i]()
                        elif self._keys_behavior[i] == HOLD:
                            self._pressed[k] = False

                elif event.type == pygame.KEYDOWN:
                    if (k := event.key) in self.key_binding:
                        i = self._key_binding[k]

                        if self._keys_behavior[i] == PRESSED:
                            self._actions[i]()
                        elif self._keys_behavior[i] == HOLD:
                            self._pressed[k] = True

                for k in self._pressed:
                    if self._pressed[k]:
                        i = self._key_binding[k]
                        self._actions[i]()

            self._clock.tick(self._fps)
            pygame.display.flip()
        self.quit()



from collections import namedtuple
import random
import math as m
from typing import Iterable, Union, Type
import numpy as np

__all__ = ["Vector"]

# Floating point precision for vectors.
EPSILON = 1e-8

Point = namedtuple("Point", ["x", "y", "z"])
Point.__new__.__defaults__ = (0, 0, 0)


class Vector(np.ndarray):
    """
    Vector
    ======
    provides :
    1. Vector support for ``python 3.9`` and upper in two or three dimensional space
    2. Fast standards operations using numpy
    3. Full random support plus ndarray and p5 compatibility

    A Vector -- specifically an Euclidean (or geometric) vector -- in
    two or three dimensional space is a geometric entity that has some
    magnitude (or length) and a direction.

    Examples
    --------
        >>> zero = Vector()
        >>> zero
        Vector(0.00, 0.00, 0.00)

        >>> vec_2d = Vector(3, 4)
        >>> vec_2d
        Vector(3.00, 4.00, 0.00)

        >>> vec_3d = Vector(2, 3, 4)
        >>> vec_3d
        Vector(2.00, 3.00, 4.00)

    Parameters
    ----------
        x : int | float
            the x-component of the vector
        y : int | float
            the y-component of the vector
        z : int | float
            the z-component of the vector
    """
    def __new__(cls, *args: Union[int, float]) -> "Vector":
        """
        new Vector instance

        Parameters
        ----------
            args : tuple[Union[int, float]]
                coordinates of the Vector, unnamed

        Returns
        -------
            Vector : new vector entity
        """
        assert not (n := len(args)) >= 4, "illegal dimension"
        (a := [e for e in args]).extend([0 for _ in range(3 - n)])
        obj = np.asarray(a, dtype=np.float64).view(cls)
        return obj

    def __eq__(self, other: "Vector") -> bool:
        if self.shape == other.shape:
            return bool(np.all(np.absolute(self - other) <= EPSILON))
        return False

    def __ne__(self, other: "Vector") -> bool:
        if self.shape == other.shape:
            return bool(np.any(np.absolute(self - other) > EPSILON))
        return True

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        fvalues = (class_name, self.x, self.y, self.z)
        return "{}({:.2f}, {:.2f}, {:.2f})".format(*fvalues)

    def __add__(self, other: "Vector") -> "Vector":
        return super().__add__(other)

    def __sub__(self, other: "Vector") -> "Vector":
        return super().__sub__(other)

    def __mul__(self, other: Union[float, int, "Vector"]) -> "Vector":
        return super().__mul__(other)

    def __rmul__(self, other: Union[float, int, "Vector"]) -> "Vector":
        return super().__rmul__(other)

    def __matmul__(self, other: "Vector") -> float:
        return self.dot(other)

    def __rmatmul__(self, other: "Vector") -> float:
        return self.dot(other)

    def __truediv__(self, other: Union[float, int, "Vector"]) -> "Vector":
        return super().__truediv__(other)

    def __floordiv__(self, other: Union[float, int, "Vector"]) -> "Vector":
        return super().__floordiv__(other)

    def __rtruediv__(self, other: Union[float, int, "Vector"]) -> "Vector":
        return super().__truediv__(other)

    def __rfloordiv__(self, other: Union[float, int, "Vector"]) -> "Vector":
        return super().__floordiv__(other)

    @property
    def x(self) -> float:
        """
        the x-component of the point
        """
        return self[0]

    @x.setter
    def x(self, value: float) -> None:
        self[0] = value

    @property
    def y(self) -> float:
        """
        the y-component of the point
        """
        return self[1]

    @y.setter
    def y(self, value: float) -> None:
        self[1] = value

    @property
    def z(self) -> float:
        """
        the z-component of the point
        """
        return self[2]

    @z.setter
    def z(self, value: float) -> None:
        self[2] = value

    def __getitem__(self, key: Union[int,
                                     slice]) -> Union[int, float, np.ndarray]:

        if isinstance(key, slice):
            return np.array([self[i] for i in range(*key.indices(len(self)))])
        return super().__getitem__(key)

    def __setitem__(self, key: Union[int, slice],
                    value: Union[int, float, list[Union[int, float]]]) -> None:
        if isinstance(key, slice):
            for i, v in zip(range(*key.indices(len(self))), value):
                self[i] = v
        else:
            super().__setitem__(key, value)

    def __delitem__(self, key: Union[int, slice]) -> None:
        if isinstance(key, slice):
            for i in range(*key.indices(len(self))):
                del self[i]
        else:
            super().__delitem__(key)

    def getCoord(self, get: str = "xyz") -> np.ndarray:
        """
        Returns the coordinates of a Vector in a numpy array

        Parameters
        ----------
            get : str, (optional)
                an ordered string of "x"s, "y"s, "z"s
                defaults to "xyz"

        Returns
        -------
            np.ndarray : coordinates
        """
        l = {"x": 0, "y": 1, "z": 2, "X": 0, "Y": 1, "Z": 2}
        keys = l.keys()
        get = [e for e in get if e in keys]
        return np.array([self[l[e]] for e in get], dtype=np.float64)

    def setCoord(self,
                 x: float = None,
                 y: float = None,
                 z: float = None) -> None:
        """
        Modifies the coordinates of the current Vector

        Parameters
        ----------
            x : float, (optional)
                the x-component
            y : float, (optional)
                the y-component
            z : float, (optional)
                the z-component
        """
        if x is None:
            x = self.x
        if y is None:
            y = self.y
        if z is None:
            z = self.z
        self.x, self.y, self.z = x, y, z

    def lerp(self, other: "Vector", amount: float) -> "Vector":
        """
        Linearly interpolate from one point to another

        Parameters
        ----------
            other : Vector
                Point to be interpolate to
            amount: float
                Amount by which to interpolate.

        Returns
        -------
            Vector : obtained by linearly interpolating this
                vector to the other vector by the given amount
        """
        x, y, z = self + amount * (other-self)
        return self.__class__(x, y, z)

    def rounded(self, ndigits: int = 0) -> "Vector":
        """
        Returns a Vector whom cordinates has been rounded with the given precision

        Parameters
        ----------
            ndigits : int
                number of digits to keep, may be negative

        Returns
        -------
            Vector : new Vector
        """
        x, y, z = self
        return self.__class__(round(x, ndigits), round(y, ndigits),
                              round(z, ndigits))

    def round(self, ndigits: int = 0) -> None:
        """
        Modifies the Vector and rounds its coordinates\\
        Does not return anything

        Parameters
        ----------
            ndigits : int
                number of digits to keep, may be negative
        """
        self.x, self.y, self.z = tuple(map(lambda x: round(x, ndigits), self))

    def inverted(self) -> "Vector":
        """
        Returns a Vector whom coordinates have been inverted, i.e. raised to power -1\\
        you might want to call ``catch_inf`` afterwards

        Returns
        -------
            Vector : new Vector
        """
        return self.__class__(*1 / self)

    def invert(self) -> None:
        """
        Modifies the Vector and invert each coordinate\\
        deal with possible division by 0 warning
        """
        def invert(x: float) -> float:
            return (1 / x, x)[abs(x) < EPSILON]

        self.x, self.y, self.z = tuple(map(invert, self))

    def catch_inf(self) -> None:
        """
        Modifies the Vector and change inf coordinates to 0
        """
        def catch(x: float) -> float:
            return (0, x)[bool(x < np.inf)]

        self.x, self.y, self.z = tuple(map(catch, self))

    def cross(self, other: "Vector") -> "Vector":
        """
        Return the cross product of the two vectors

        Examples
        --------
            >>> i = Vector(1, 0, 0)
            >>> j = Vector(0, 1, 0)
            >>> i.cross(j)
            Vector(0.00, 0.00, 1.00)

        Parameters
        ----------
            other : Vector

        Returns
        -------
            Vector : The vector perpendicular to both ``self`` and ``other``
                i.e., the vector obtained by taking the cross product of
                ``self`` and ``other``
        """
        x, y, z = np.cross(self, other)
        return self.__class__(x, y, z)

    def dot(self, other: "Vector") -> float:
        """
        Computes the dot product of two vectors

        Examples
        --------
            >>> p = Vector(2, 3, 6)
            >>> q = Vector(3, 4, 5)
            >>> p.dot(q)
            48
            >>> p @ q
            48

        Parameters
        ----------
            other : Vector

        Returns
        -------
            float : the dot product of the two vectors
        """
        return np.dot(self, other)

    @classmethod
    def random(cls,
               v1: float,
               v2: float,
               size: int = 3,
               dtype: Type[Union[float, int]] = float) -> "Vector":
        """
        Creates a random generated Vector\\
        both ``v1`` and ``v2`` are included

        Examples
        --------
            >>> v1 = Vector.random(3, 10, size=2, dtype=int)
            >>> v2 = Vector.random(3, 10, size=2, dtype=float)
            >>> print(v1, v2, sep="\\n")
            ... # Vector(9.00, 7.00, 0.00)
            ... # Vector(7.88, 8.91, 0.00)

        Parameters
        ----------
            v1 : float or int
                the minimum value
            v2 : float or int
                the maximum value
            size : int, (optional)
                the number of random-generated coordinates : min=0 | max=3
                defaults to 3
            dtype : type, (optional)
                the type of data : float or int
                defaults to float

        Returns
        -------
            Vector : a new Vector
        """
        assert not ((size := abs(size)) >= 4), "please enter valid size"
        f = random.uniform if dtype is float else random.randint
        cast = float if dtype is float else int
        x, y, z = 0, 0, 0
        if size >= 1:
            x = f(cast(v1), cast(v2))
        if size >= 2:
            y = f(cast(v1), cast(v2))
        if size == 3:
            z = f(cast(v1), cast(v2))
        return cls(x, y, z)

    @classmethod
    def random2d(cls, mag: float = 1) -> "Vector":
        """
        Generates a random 2d vector with an optional desired magnitude
        """
        x, y = 2 * (np.random.random(2) - 0.5)
        vec = cls(x, y)
        vec.magnitude = mag
        return vec

    @classmethod
    def random3d(cls, mag: float = 1) -> "Vector":
        """
        Generates a random 3d vector with an optional desired magnitude
        """
        x, y, z = 2 * (np.random.random(3) - 0.5)
        vec = cls(x, y, z)
        vec.magnitude = mag
        return vec

    def copy(self) -> "Vector":
        """
        Return a copy of the current point

        Returns
        -------
            Vector : a copy of the current point
        """
        x, y, z = self
        return self.__class__(x, y, z)

    @property
    def magnitude(self) -> float:
        """
        The magnitude of the vector

        Examples
        --------
            >>> p = Vector(2, 3, 6)
            >>> p.magnitude
            7.0

            >>> abs(p)
            7.0

            >>> p.magnitude = 14
            >>> p
            Vector(4.00, 6.00, 12.00)

            >>> p.normalize()
            >>> print(p)
            Vector(0.29, 0.43, 0.86)
        """
        return m.sqrt(np.dot(self, self))

    @magnitude.setter
    def magnitude(self, value: float) -> None:
        if not (m := self.magnitude) == 0:
            self *= abs(value) / m

    @property
    def magnitude_sq(self) -> float:
        """
        The squared magnitude of the vector
        """
        return np.dot(self, self)

    @magnitude_sq.setter
    def magnitude_sq(self, value: float) -> None:
        self.magnitude = m.sqrt(value)

    def __abs__(self) -> float:
        return self.magnitude

    def normalize(self) -> None:
        """
        Sets the magnitude of the vector to 1
        """
        assert not self.magnitude == 0, "vector has magnitude 0, can't normalize"
        self.magnitude = 1

    def normalized(self) -> "Vector":
        """
        Returns a copy of the current vector with a magnitude of 1
        """
        assert not self.magnitude == 0, "vector has magnitude 0, can't normalize"
        other = self.copy()
        other.magnitude = 1
        return other

    def limit(self, upper: float = None, lower: float = None) -> None:
        """
        Keeps the vector magnitude under or above a given limit
        """
        m = self.magnitude
        if lower is None:
            lower = m
        if upper is None:
            upper = m

        if m < lower:
            self.magnitude = lower
        elif m > upper:
            self.magnitude = upper

    def limited(self, upper: float = None, lower: float = None) -> "Vector":
        """
        Returns a new vector whom magnitude has been keeped under or above a given limit
        """
        other = self.copy()
        m = other.magnitude
        if lower is None:
            lower = m
        if upper is None:
            upper = m

        if m < lower:
            other.magnitude = lower
        elif m > upper:
            other.magnitude = upper
        return other

    def distance(self, other: "Vector") -> float:
        """
        Return the distance between two points

        Parameters
        ----------
            other : Vector
                other vector

        Returns
        -------
            float : The distance between the current point and the given point
        """
        return m.sqrt(sum((v := (self - other)) * v))

    def distance_sq(self, other: "Vector") -> float:
        """
        Return the squared distance between two points

        Parameters
        ----------
            other : Vector
                other vector

        Returns
        -------
            float : The squared distance between the current point and the given point
        """
        return sum((v := (self - other)) * v)

    @property
    def angle(self) -> float:
        """
        The angle of rotation of the vector (in radians)\\
        This attribute isn't available for three dimensional vectors

        Examples
        --------
            >>> from math import pi, isclose
            >>> p = Vector(1, 0, 0)
            >>> isclose(p.angle, 0)
            True

            >>> p = Vector(0, 1, 0)
            >>> isclose(p.angle, pi/2)
            True

            >>> p = Vector(1, 1)
            >>> isclose(p.angle, pi/4)
            True
            >>> p.angle = pi/2
            >>> isclose(p.angle, pi/2)
            True

            >>> p = Vector(1, 1)
            >>> isclose(p.angle, pi/4)
            True
            >>> p.rotate(pi/4)
            >>> isclose(p.angle, pi/2)
            True
        """
        assert not np.abs(
            self.z) > EPSILON, "can't compute angle for 3d vectors"
        return np.arctan2(self.y, self.x)

    @angle.setter
    def angle(self, theta: float) -> None:
        self.rotate(theta - self.angle)

    def rotate(self, theta: float) -> None:
        """
        Rotates the vector by an angle

        Parameters
        ----------
            theta : float or int
                angle in radians
        """
        x = self.x * m.cos(theta) - self.y * m.sin(theta)
        y = self.x * m.sin(theta) + self.y * m.cos(theta)
        self.x = x
        self.y = y

    def rotated(self, theta: float) -> "Vector":
        """
        Returns a new vector which has been rotated by an angle

        Parameters
        ----------
            theta : float or int
                angle in radians

        Returns
        -------
            Vector : new vector
        """
        x = self.x * m.cos(theta) - self.y * m.sin(theta)
        y = self.x * m.sin(theta) + self.y * m.cos(theta)
        return self.__class__(x, y)

    def angle_between(self, other: "Vector") -> float:
        """
        Calculate the angle between two vectors

        Examples
        --------
            >>> from math import degrees
            >>> k = Vector(0, 1)
            >>> j = Vector(1, 0)
            >>> degrees(k.angle_between(j))
            90.0

        Parameters
        ----------
            other : Vector

        Returns
        -------
            float : The angle between two given vectors (in radians)
        """
        return np.arccos(
            (np.dot(self, other)) / (self.magnitude * other.magnitude))

    @classmethod
    def from_angle(cls, angle: float) -> "Vector":
        """
        Return a new unit vector with the given angle

        Parameters
        ----------
        angle : float
            Angle to be used to create the vector (in radians)
        """
        vec = cls.random2d(mag=1)
        vec.angle = angle
        return vec

    # aliases
    __str__ = __repr__
    mag = norm = magnitude
    mag2 = magnitude_sq
    dist = distance
    dist_sq = distance_sq
    heading = angle
    u = phi = p = x
    v = theta = q = y
    w = psi = r = z



from typing import Union
from ..data import *

import math as m

__all__ = ["ScrollBar"]

import difflib
import pygame


def _map(x: float, x0: float, x1: float, y0: float, y1: float) -> float:
    """
    linear interpolation
    """
    return (y0 * (x1-x) + y1 * (x-x0)) / (x1-x0)


class ScrollBar:
    """
    Pygame ScrollBar
    =============
    created by ``Renderer``
    """
    def __init__(self,
                 renderer,
                 yrange: tuple[int, int],
                 color1: Union[tuple[int, int, int], int, str] = None,
                 color2: Union[tuple[int, int, int], int, str] = None) -> None:
        """
        new ScrollBar instance
        
        Parameters
        ----------
            renderer : Renderer
                main renderer
            yrange : tuple[int, int]
                minimum and maximum y view point
            color1 : Union[tuple[int, int, int], int, str], (optional)
                color of the scrollbar item
                defaults to None
            color2 : Union[tuple[int, int, int], int, str], (optional)
                color of the background when activated
                defaults to None
        """
        self.has_error = False
        self._renderer = renderer

        self._min_val = min(yrange)
        self._max_val = max(yrange)
        self._value = self._min_val

        self._y_pin = 0
        self._pinned = False

        # if self._max_val - self._min_val <= self._renderer.win_height:
        #     warn(
        #         f"ERROR [active scrollbar] : yrange must exceed the window height, scrollbar was not created")
        #     self.has_error = True

        h = self._renderer.win_height
        r = self._max_val - self._min_val
        self._height = 30 + (h-30) / m.sqrt(r + 1)

        color1 = 155 if color1 is None else color1
        if isinstance(color1, tuple) and len(color1) == 3:
            self._color1 = color1
        elif isinstance(color1, int):
            self._color1 = color1, color1, color1
        elif isinstance(color1, str):
            try:
                self._color1 = COLORS[color1.lower()]
            except KeyError:
                close = difflib.get_close_matches(color1.lower(),
                                                  COLORS.keys(),
                                                  n=1,
                                                  cutoff=.5)[0]
                warn(
                    f"ERROR [active scrollbar] : {color1} is not a valid color name, using closest match {close} instead"
                )
                self._color1 = COLORS[close]
        else:
            warn(
                f"ERROR [active scrollbar] : wrong color1 parameter, scrollbar was not created"
            )
            self.has_error = True

        color2 = 50 if color2 is None else color2
        if isinstance(color2, tuple) and len(color2) == 3:
            self._color2 = color2
        elif isinstance(color2, int):
            self._color2 = color2, color2, color2
        elif isinstance(color2, str):
            try:
                self._color2 = COLORS[color2.lower()]
            except KeyError:
                close = difflib.get_close_matches(color2.lower(),
                                                  COLORS.keys(),
                                                  n=1,
                                                  cutoff=.5)[0]
                warn(
                    f"ERROR [active scrollbar] : {color2} is not a valid color name, using closest match {close} instead"
                )
                self._color2 = COLORS[close]
        else:
            warn(
                f"ERROR [active scrollbar] : wrong color2 parameter, scrollbar was not created"
            )
            self.has_error = True

        self._is_hidden = False

        self._tick_count = 1
        self._max_ticks = 0
        self._is_playing = False
        self._is_active = False

    @property
    def min_val(self) -> int:
        """
        gets scrollbar minimum value
        """
        return self._min_val

    @property
    def max_val(self) -> int:
        """
        gets scrollbar maximum value
        """
        return self._max_val

    @property
    def value(self) -> float:
        """
        gets scrollbar current value
        """
        return self._value

    @property
    def height(self) -> float:
        """
        gets scrollbar height
        """
        return self._height

    @property
    def ypos(self) -> float:
        """
        gets desired position on the window based on current value
        """
        h = self._renderer.win_height
        y = _map(self.value, self.min_val, self.max_val, 0, h - self._height)
        return y

    @property
    def is_active(self) -> bool:
        """
        gets current state of the scrollbar
        """
        return self._is_active

    @property
    def is_playing(self) -> bool:
        """
        gets current state of animation of the scrollnar
        """
        return self._is_playing

    @property
    def is_hidden(self) -> bool:
        """
        gets current display state of the scrollbar
        """
        return self._is_hidden

    @is_hidden.setter
    def is_hidden(self, value: bool) -> None:
        """
        forces the display state of the scrollbar
        """
        self._is_hidden = value

    @property
    def color1(self) -> tuple[int, int, int]:
        """
        gets current color1
        """
        return self._color1

    @color1.setter
    def color1(self, color1: Union[tuple[int, int, int], int, str]) -> None:
        """
        sets scrollbar color1
        """
        if isinstance(color1, tuple) and len(color1) == 3:
            self._color1 = color1
        elif isinstance(color1, int):
            self._color1 = color1, color1, color1
        elif isinstance(color1, str):
            try:
                self._color1 = COLORS[color1.lower()]
            except KeyError:
                close = difflib.get_close_matches(color1.lower(),
                                                  COLORS.keys(),
                                                  n=1,
                                                  cutoff=.5)[0]
                warn(
                    f"ERROR [active scrollbar] : {color1} is not a valid color name, using closest match {close} instead"
                )
                self._color1 = COLORS[close]
        else:
            warn(
                f"ERROR [active scrollbar] : wrong color1 parameter, nothing changed"
            )

    @property
    def color2(self) -> tuple[int, int, int]:
        """
        gets current color2
        """
        return self._color2

    @color2.setter
    def color2(self, color2) -> None:
        """
        sets scrollbar color2
        """
        if isinstance(color2, tuple) and len(color2) == 3:
            self._color2 = color2
        elif isinstance(color2, int):
            self._color2 = color2, color2, color2
        elif isinstance(color2, str):
            try:
                self._color2 = COLORS[color2.lower()]
            except KeyError:
                close = difflib.get_close_matches(color2.lower(),
                                                  COLORS.keys(),
                                                  n=1,
                                                  cutoff=.5)[0]
                warn(
                    f"ERROR [active scrollbar] : {color2} is not a valid color name, using closest match {close} instead"
                )
                self._color2 = COLORS[close]
        else:
            warn(
                f"ERROR [active scrollbar] : wrong color2 parameter, nothing changed"
            )

    def hide(self) -> None:
        """
        hides the scrollbar (does not display it automatically on the screen)\\
        scrollbar actions become unaccessible\\
        opposite method is ``reveal``
        """
        if self._is_hidden:
            warn(
                f"WARNING [active scrollbar] : scrollbar is already hidden, nothing changed"
            )
            return
        self._is_hidden = True

    def reveal(self) -> None:
        """
        reveals the scrollbar back (displays it on the screen)\\
        scrollbar actions become accessible again\\
        opposite method is ``hide``
        """
        if not self._is_hidden:
            warn(
                f"WARNING [active scrollbar] : scrollbar is not hidden, nothing changed"
            )
            return
        self._is_hidden = False

    def set_max_ticks(self, sec: float = .1) -> None:
        """
        sets max tick for animation base on fps
        """
        fps = self._renderer.fps
        self._max_ticks = round(fps * sec)

    def collide(self, pos: tuple[int, int]) -> bool:
        """
        if mouse is inside
        """
        if self._pinned:
            return True
        return (self._renderer.win_width - 15 <= pos[0] <= self._renderer.win_width)\
               and (self.ypos <= pos[1] <= self.ypos + self.height)

    def get_pin(self) -> float:
        """
        pin y position
        """
        return self._y_pin

    def is_pinned(self) -> bool:
        """
        if the scrollbar has been pinned
        """
        return self._pinned

    def set_pin(self, pos: tuple[int, int]) -> None:
        """
        set the pin
        """
        self._pinned = True
        self._y_pin = pos[1] - self.ypos

    def unpin(self) -> None:
        """
        unpin the scrollbar
        """
        self._pinned = False
        self._y_pin = 0

    def animate(self) -> None:
        """
        go trough animation when hoovering\\
        does one frame of animation based on tick_count\\
        ends animation if needed
        """
        self._tick_count += self._is_playing
        if self._tick_count >= self._max_ticks - 1:
            self._max_ticks = 0
            self._tick_count = 1
            self._is_playing = False

    def deactivate(self) -> None:
        """
        deactivate the scrollbar (hoover effect)
        """
        if (not self._is_active) or self._is_playing:
            warn(
                f"WARNING [active scrollbar] : scrollbar is already non active or is currently animated, nothing changed"
            )
            return
        self._is_active = False
        self._is_playing = True
        self.set_max_ticks()

    def activate(self) -> None:
        """
        activate the scrollbar (hoover effect)
        """
        if self._is_active or self._is_playing:
            warn(
                f"WARNING [active scrollbar] : scrollbar is already active or is currently animated, nothing changed"
            )
            return
        self._is_active = True
        self._is_playing = True
        self.set_max_ticks()

    def set_value_by_y(self, ypos: float) -> None:
        """
        sets the value of the scrollbar depending on a top y position
        """
        v = _map(ypos, 0, self._renderer.win_height - self.height,
                 self.min_val, self.max_val)
        self._value = v
        if self._value < self.min_val:
            self._value = self._min_val
        if self._value > self.max_val:
            self._value = self.max_val

    def scroll_up(self, amount: float = .05) -> None:
        """
        scrolls up by some amount (as a fraction)
        """
        total = self.max_val - self.min_val
        amt = amount * total
        self._value -= amt

        if self._value < self.min_val:
            self._value = self._min_val

    def scroll_down(self, amount: float = .05) -> None:
        """
        scrolls down by some amount (as a fraction)
        """
        total = self.max_val - self.min_val
        amt = amount * total
        self._value += amt

        if self._value > self.max_val:
            self._value = self.max_val

    def get_y_translation(self) -> float:
        """
        gets translation for the renderer
        """
        return self.value

    def update_state(self, pos: tuple[int, int]) -> None:
        """
        updates state of the active scrollbar
        """
        renderer = self._renderer

        if self.is_active:
            if pos[0] < renderer.win_width - 15:
                self.deactivate()

        elif not self.is_active:
            if pos[0] >= renderer.win_width - 5:
                self.activate()

    def draw(self) -> None:
        """
        draws scrollbar on the screen\\
        different render method depending on the mouse pos
        """
        window = self._renderer._window

        bg: tuple[int, int, int] = self._renderer.win_bg
        width: int = self._renderer.win_width
        height: int = self._renderer.win_height

        if self.is_active:
            if not self.is_playing:
                pygame.draw.rect(window, self.color2,
                                 ((width - 15, 0), (15, height)), 0)
                pygame.draw.rect(window, self.color1,
                                 ((width - 15, self.ypos), (15, self.height)),
                                 0)

            else:
                w = _map(self._tick_count, self._max_ticks, 0, 15, 5)
                r = _map(self._tick_count, self._max_ticks, 0, self.color2[0],
                         bg[0])
                g = _map(self._tick_count, self._max_ticks, 0, self.color2[1],
                         bg[1])
                b = _map(self._tick_count, self._max_ticks, 0, self.color2[2],
                         bg[2])

                pygame.draw.rect(window, (r, g, b),
                                 ((width - 15, 0), (15, height)), 0)
                pygame.draw.rect(window, self.color1,
                                 ((width - w, self.ypos), (w, self.height)), 0)

            self.animate()

        else:
            if not self.is_playing:
                pygame.draw.line(window, self.color1, (width - 2, self.ypos),
                                 (width - 2, self.ypos + self.height), 1)
                for i in range(2):
                    pygame.draw.line(
                        window, self.color1, (width - 2 + i, self.ypos + i),
                        (width - 2 + i, self.ypos + self.height - i), 1)
                    pygame.draw.line(
                        window, self.color1, (width - 2 - i, self.ypos + i),
                        (width - 2 - i, self.ypos + self.height - i), 1)

            else:
                w = _map(self._tick_count, 0, self._max_ticks, 15, 5)
                r = _map(self._tick_count, 0, self._max_ticks, self.color2[0],
                         bg[0])
                g = _map(self._tick_count, 0, self._max_ticks, self.color2[1],
                         bg[1])
                b = _map(self._tick_count, 0, self._max_ticks, self.color2[2],
                         bg[2])

                pygame.draw.rect(window, (r, g, b),
                                 ((width - 15, 0), (15, height)), 0)
                pygame.draw.rect(window, self.color1,
                                 ((width - w, self.ypos), (w, self.height)), 0)

            self.animate()
